# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
from gettext import ngettext
import threading

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from komikku.library.categories_list import CategoriesList
from komikku.library.thumbnail import Thumbnail
from komikku.models import Category
from komikku.models import CategoryVirtual
from komikku.models import create_db_connection
from komikku.models import Manga
from komikku.models import Settings
from komikku.models import update_rows


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/library.ui')
class LibraryPage(Adw.NavigationPage):
    __gtype_name__ = 'LibraryPage'

    left_button = Gtk.Template.Child('left_button')
    categories_togglebutton = Gtk.Template.Child('categories_togglebutton')
    filters_button = Gtk.Template.Child('filters_button')
    title = Gtk.Template.Child('title')
    search_button = Gtk.Template.Child('search_button')
    menu_button = Gtk.Template.Child('menu_button')

    searchbar = Gtk.Template.Child('searchbar')
    searchbar_separator = Gtk.Template.Child('searchbar_separator')
    search_entry = Gtk.Template.Child('searchentry')
    overlaysplitview = Gtk.Template.Child('overlaysplitview')
    stack = Gtk.Template.Child('stack')

    categories_stack = Gtk.Template.Child('categories_stack')
    categories_listbox = Gtk.Template.Child('categories_listbox')
    categories_edit_mode_buttonbox = Gtk.Template.Child('categories_edit_mode_buttonbox')
    categories_edit_mode_cancel_button = Gtk.Template.Child('categories_edit_mode_cancel_button')
    categories_edit_mode_ok_button = Gtk.Template.Child('categories_edit_mode_ok_button')

    activity_progressbar = Gtk.Template.Child('activity_progressbar')
    flowbox = Gtk.Template.Child('flowbox')
    selection_mode_actionbar = Gtk.Template.Child('selection_mode_actionbar')
    selection_mode_menubutton = Gtk.Template.Child('selection_mode_menubutton')

    start_page_progressbar = Gtk.Template.Child('start_page_progressbar')
    start_page_logo_image = Gtk.Template.Child('start_page_logo_image')
    start_page_title_label = Gtk.Template.Child('start_page_title_label')
    start_page_discover_button = Gtk.Template.Child('start_page_discover_button')

    page = None
    populating = False
    selected_filters = []
    selection_mode = False
    selection_mode_range = False
    selection_mode_last_thumbnail_index = None
    thumbnails_cover_size = None

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window
        self.builder = window.builder
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/main.xml')
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/library_selection_mode.xml')

        self.selected_filters = Settings.get_default().library_selected_filters

        self.connect('shown', self.on_shown)

        # Header bar
        self.left_button.connect('clicked', self.on_left_button_clicked)
        self.filters_button.connect('clicked', self.on_filters_button_clicked)
        if self.selected_filters:
            self.filters_button.add_css_class('accent')
        self.menu_button.set_menu_model(self.builder.get_object('menu'))
        # Focus is lost after showing popover submenu (bug?)
        self.menu_button.get_popover().connect('closed', lambda _popover: self.menu_button.grab_focus())

        # Search
        self.search_entry.connect('activate', self.on_search_entry_activated)
        self.search_entry.connect('changed', self.search)
        self.searchbar.bind_property(
            'search-mode-enabled', self.search_button, 'active',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.bind_property(
            'search-mode-enabled', self.searchbar_separator, 'visible',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )
        self.searchbar.connect_entry(self.search_entry)
        self.searchbar.set_key_capture_widget(self.window)

        # Filters dialog
        self.filters_dialog = Adw.PreferencesDialog.new()
        self.filters_dialog.set_title(_('Filters'))
        self.filters_dialog.props.presentation_mode = Adw.DialogPresentationMode.BOTTOM_SHEET
        self.init_filters_dialog()

        # Overlay split view (provide overlay sidebar)
        self.overlaysplitview.connect('notify::show-sidebar', self.on_overlaysplitview_revealed)
        self.overlaysplitview.bind_property(
            'show-sidebar', self.categories_togglebutton, 'active',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )

        self.categories_list = CategoriesList(self)

        # Thumbnails Flowbox
        self.flowbox.set_valign(Gtk.Align.START)
        self.flowbox.connect('child-activated', self.on_manga_thumbnail_activated)
        self.flowbox.connect('selected-children-changed', self.update_title)
        self.flowbox.connect('unselect-all', self.leave_selection_mode)

        # Selection mode ActionBar
        self.selection_mode_menubutton.set_menu_model(self.builder.get_object('menu-library-selection-mode'))

        # Gestures for multi-selection mode
        self.window.controller_key.connect('key-pressed', self.on_key_pressed)

        self.gesture_click = Gtk.GestureClick.new()
        self.gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_click.set_button(3)
        self.gesture_click.connect('released', self.on_manga_thumbnail_right_click)
        self.flowbox.add_controller(self.gesture_click)

        self.gesture_long_press = Gtk.GestureLongPress.new()
        self.gesture_long_press.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_long_press.set_touch_only(False)
        self.gesture_long_press.connect('pressed', self.on_gesture_long_press_activated)
        self.flowbox.add_controller(self.gesture_long_press)

        self.window.updater.connect('manga-updated', self.on_manga_updated)

        def _filter(thumbnail):
            manga = thumbnail.manga
            selected_category = Settings.get_default().selected_category

            if selected_category != CategoryVirtual.ALL:
                if selected_category == CategoryVirtual.UNCATEGORIZED:
                    # Virtual category 'Uncategorized' is selected
                    ret = not manga.categories
                else:
                    # Categorized
                    ret = selected_category in manga.categories
            else:
                # All
                ret = True

            if ret:
                term = self.search_entry.get_text().lower()

                # Search in name
                ret = term in manga.name.lower()

                # Search in server name
                ret = ret or term in manga.server.name.lower()

                # Search in genres (exact match)
                ret = ret or term in [genre.lower() for genre in manga.genres]

                # Optional menu filters
                if ret and 'downloaded' in self.selected_filters:
                    ret = manga.nb_downloaded_chapters > 0
                if ret and 'unread' in self.selected_filters:
                    ret = manga.nb_unread_chapters > 0
                if ret and 'recents' in self.selected_filters:
                    ret = manga.nb_recent_chapters > 0

            if not ret and thumbnail._selected:
                # Unselect thumbnail if it's selected
                self.flowbox.unselect_child(thumbnail)
                thumbnail._selected = False

            thumbnail._filtered = not ret

            return ret

        def _sort(thumbnail1, thumbnail2):
            """
            This function gets two children and has to return:
            - a negative integer if the firstone should come before the second one
            - zero if they are equal
            - a positive integer if the second one should come before the firstone
            """
            manga1 = thumbnail1.manga
            manga2 = thumbnail2.manga

            sort_order = Settings.get_default().library_sort_order
            if sort_order == 'latest-read-desc':
                if manga1.last_read.timestamp() > manga2.last_read.timestamp():
                    return -1

                if manga1.last_read.timestamp() < manga2.last_read.timestamp():
                    return 1

            elif sort_order == 'latest-updated-desc':
                if manga1.last_update is None or manga2.last_update is None:
                    if manga1.last_update is not None and manga2.last_update is None:
                        return -1
                    if manga1.last_update is None and manga2.last_update is not None:
                        return 1

                    # Both are None, fall back on alphabetical sorting (ascendant)
                    if manga1.name > manga2.name:
                        return 1
                    if manga1.name < manga2.name:
                        return -1

                    return 0

                if manga1.last_update.timestamp() > manga2.last_update.timestamp():
                    return -1

                if manga1.last_update.timestamp() < manga2.last_update.timestamp():
                    return 1

            elif sort_order == 'alphanum-asc':
                if manga1.name > manga2.name:
                    return 1

                if manga1.name < manga2.name:
                    return -1

            return 0

        self.flowbox.set_filter_func(_filter)
        self.flowbox.set_sort_func(_sort)

        self.window.navigationview.add(self)

    def add_accelerators(self):
        self.window.application.set_accels_for_action('app.library.update', ['<Primary>r'])
        self.window.application.set_accels_for_action('app.library.toggle-categories', ['F9'])

    def add_actions(self):
        add_action = Gio.SimpleAction.new('add', None)
        add_action.connect('activate', self.on_left_button_clicked)
        self.window.application.add_action(add_action)

        # Menu actions
        update_action = Gio.SimpleAction.new('library.update', None)
        update_action.connect('activate', self.update_all)
        self.window.application.add_action(update_action)

        variant = GLib.Variant.new_string(Settings.get_default().library_sort_order)
        self.sort_order_action = Gio.SimpleAction.new_stateful('library.sort-order', variant.get_type(), variant)
        self.sort_order_action.connect('activate', self.on_sort_order_changed)
        self.window.application.add_action(self.sort_order_action)

        download_manager_action = Gio.SimpleAction.new('library.download-manager', None)
        download_manager_action.connect('activate', self.open_download_manager)
        self.window.application.add_action(download_manager_action)

        categories_editor_action = Gio.SimpleAction.new('library.categories-editor', None)
        categories_editor_action.connect('activate', self.open_categories_editor)
        self.window.application.add_action(categories_editor_action)

        history_action = Gio.SimpleAction.new('library.history', None)
        history_action.connect('activate', self.open_history)
        self.window.application.add_action(history_action)

        # ActionBar actions in selection mode
        update_selected_action = Gio.SimpleAction.new('library.update-selected', None)
        update_selected_action.connect('activate', self.update_selected)
        self.window.application.add_action(update_selected_action)

        delete_selected_action = Gio.SimpleAction.new('library.delete-selected', None)
        delete_selected_action.connect('activate', self.delete_selected)
        self.window.application.add_action(delete_selected_action)

        download_selected_action = Gio.SimpleAction.new('library.download-selected', None)
        download_selected_action.connect('activate', self.download_selected)
        self.window.application.add_action(download_selected_action)

        mark_selected_read_action = Gio.SimpleAction.new('library.mark-selected-read', None)
        mark_selected_read_action.connect('activate', self.toggle_selected_read_status, 1)
        self.window.application.add_action(mark_selected_read_action)

        mark_selected_unread_action = Gio.SimpleAction.new('library.mark-selected-unread', None)
        mark_selected_unread_action.connect('activate', self.toggle_selected_read_status, 0)
        self.window.application.add_action(mark_selected_unread_action)

        edit_categories_selected_action = Gio.SimpleAction.new('library.edit-categories-selected', None)
        edit_categories_selected_action.connect('activate', self.edit_categories_selected)
        self.window.application.add_action(edit_categories_selected_action)

        select_all_action = Gio.SimpleAction.new('library.select-all', None)
        select_all_action.connect('activate', self.select_all)
        self.window.application.add_action(select_all_action)

        # Shortcuts
        toggle_categories_action = Gio.SimpleAction.new('library.toggle-categories', None)
        toggle_categories_action.connect(
            'activate', lambda _action, _param: self.overlaysplitview.set_show_sidebar(
                not self.overlaysplitview.get_show_sidebar()
            )
        )
        self.window.application.add_action(toggle_categories_action)

    def compute_thumbnails_cover_size(self):
        default_width = Thumbnail.default_width
        default_height = Thumbnail.default_height

        container_width = self.window.get_width() or self.window.props.default_width

        if container_width / default_width != container_width // default_width:
            nb = container_width // default_width + 1
            width = container_width // nb
        else:
            width = default_width
        width -= Thumbnail.padding * 2 + Thumbnail.margin * 2

        height = (width * default_height) // default_width

        self.thumbnails_cover_size = (width, height)

    def delete_mangas(self, mangas):
        def confirm_callback():
            # Stop Downloader & Updater
            self.window.downloader.stop()
            self.window.updater.stop()

            GLib.idle_add(do_delete)

        def do_delete():
            if self.window.downloader.running or self.window.updater.running:
                return GLib.SOURCE_CONTINUE

            db_conn = create_db_connection()

            # Safely delete mangas in DB
            for manga in mangas:
                manga.delete(db_conn=db_conn)

            # Restart Downloader & Updater
            self.window.downloader.start()
            self.window.updater.start()

            # Finally, update library
            for manga in mangas:
                self.remove_thumbnail(manga)

            if self.window.page == 'card':
                self.window.navigationview.pop()
            else:
                self.leave_selection_mode()

            if not self.flowbox.get_first_child():
                # Library is now empty
                self.populate()

        if self.window.page == 'card':
            message = _('Are you sure you want to delete this manga?')
        else:
            message = _('Are you sure you want to delete selected mangas?')
        self.window.open_dialog(
            _('Delete?'),
            body=message,
            confirm_label=_('Delete'),
            confirm_callback=confirm_callback,
            confirm_appearance=Adw.ResponseAppearance.DESTRUCTIVE
        )

    def delete_selected(self, _action, _param):
        self.delete_mangas([thumbnail.manga for thumbnail in self.flowbox.get_selected_children()])

    def download_selected(self, _action, _param):
        def confirm_callback():
            chapters = []
            for thumbnail in self.flowbox.get_selected_children():
                for chapter in thumbnail.manga.chapters:
                    chapters.append(chapter)

            self.leave_selection_mode()

            self.window.downloader.add(chapters)
            self.window.downloader.start()

        message = _('Are you sure you want to download all chapters of selected mangas?')
        self.window.open_dialog(
            _('Download?'),
            body=message,
            confirm_label=_('Download'),
            confirm_callback=confirm_callback,
            confirm_appearance=Adw.ResponseAppearance.SUGGESTED
        )

    def edit_categories_selected(self, _action, _param):
        # Edit categories of selected mangas
        self.categories_list.set_edit_mode(True)
        self.overlaysplitview.set_show_sidebar(True)

    def enter_selection_mode(self):
        self.selection_mode = True
        self.update_headerbar_buttons()

        self.flowbox.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
        self.selection_mode_actionbar.set_revealed(True)

    def init_filters_dialog(self):
        page = Adw.PreferencesPage(title=_('Filters'))
        self.filters_dialog.add(page)

        group = Adw.PreferencesGroup()

        filters = {
            'unread': {
                'title': _('Unread'),
                'subtitle': _('Series with unread chapters'),
            },
            'downloaded': {
                'title': _('Downloaded'),
                'subtitle': _('Series with downloaded chapters'),
            },
            'recents': {
                'title': _('Recents'),
                'subtitle': _('Series with new chapters'),
            },
        }

        for name, filter in filters.items():
            row = Adw.SwitchRow(title=filter['title'])
            row.set_subtitle(filter['subtitle'])
            row.set_use_markup(True)
            row.set_active(name in self.selected_filters)
            row.connect('notify::active', self.on_filter_changed, name)
            group.add(row)

        page.add(group)

    def leave_selection_mode(self, _param=None):
        self.selection_mode = False
        self.update_headerbar_buttons()

        self.flowbox.set_selection_mode(Gtk.SelectionMode.NONE)
        for thumbnail in self.flowbox:
            thumbnail._selected = False

        self.selection_mode_actionbar.set_revealed(False)
        self.overlaysplitview.set_show_sidebar(False)

    def on_filter_changed(self, switch, _gparam, name):
        if switch.get_active():
            self.selected_filters.add(name)
        else:
            self.selected_filters.remove(name)
        Settings.get_default().library_selected_filters = self.selected_filters

        if self.selected_filters:
            self.filters_button.add_css_class('accent')
        else:
            self.filters_button.remove_css_class('accent')

        self.flowbox.invalidate_filter()

    def on_filters_button_clicked(self, _button):
        self.filters_dialog.present(self.window)

    def on_gesture_long_press_activated(self, _gesture, x, y):
        """Allow to enter in selection mode with a long press on a thumbnail"""
        if not self.selection_mode:
            self.enter_selection_mode()
        else:
            # Enter in 'Range' selection mode
            # Long press on a manga then long press on another to select everything in between
            self.selection_mode_range = True

        selected_thumbnail = self.flowbox.get_child_at_pos(x, y)
        self.on_manga_thumbnail_activated(None, selected_thumbnail)

    def on_key_pressed(self, _controller, keyval, _keycode, state):
        if self.window.page != self.props.tag:
            return Gdk.EVENT_PROPAGATE

        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        if not self.selection_mode:
            # Allow to enter in selection mode with <SHIFT>+Arrow key
            arrow_keys = (
                Gdk.KEY_Up, Gdk.KEY_KP_Up,
                Gdk.KEY_Down, Gdk.KEY_KP_Down,
                Gdk.KEY_Left, Gdk.KEY_KP_Left,
                Gdk.KEY_Right, Gdk.KEY_KP_Right
            )
            if modifiers != Gdk.ModifierType.SHIFT_MASK or keyval not in arrow_keys:
                return Gdk.EVENT_PROPAGATE

            thumbnail = self.flowbox.get_focus_child() or self.flowbox.get_first_child()
            if thumbnail is not None:
                self.enter_selection_mode()
                self.on_manga_thumbnail_activated(None, thumbnail)
        else:
            if keyval == Gdk.KEY_Escape or (modifiers == Gdk.ModifierType.ALT_MASK and keyval in (Gdk.KEY_Left, Gdk.KEY_KP_Left)):
                self.leave_selection_mode()

        return Gdk.EVENT_PROPAGATE

    def on_left_button_clicked(self, action_or_button=None, _param=None):
        if self.window.page != self.props.tag:
            return

        if not self.selection_mode:
            self.window.explorer.show()
        else:
            self.leave_selection_mode()

    def on_manga_added(self, manga):
        """Called from 'Card' when user clicks on `+ Add to Library` button"""
        db_conn = create_db_connection()
        nb_mangas = db_conn.execute('SELECT count(*) FROM mangas WHERE in_library = 1').fetchone()[0]

        if nb_mangas == 1:
            # Library was previously empty
            self.populate()
        else:
            thumbnail = Thumbnail(self, manga, *self.thumbnails_cover_size)
            self.flowbox.prepend(thumbnail)

    def on_manga_thumbnail_activated(self, _flowbox, thumbnail):
        if self.selection_mode:
            if self.selection_mode_range and self.selection_mode_last_thumbnail_index is not None:
                # Range selection mode: select all mangas between last selected manga and clicked manga
                walk_index = self.selection_mode_last_thumbnail_index
                last_index = thumbnail.get_index()

                while walk_index != last_index:
                    walk_thumbnail = self.flowbox.get_child_at_index(walk_index)
                    if walk_thumbnail and not walk_thumbnail._selected and not walk_thumbnail._filtered:
                        self.flowbox.select_child(walk_thumbnail)
                        walk_thumbnail._selected = True

                    if walk_index < last_index:
                        walk_index += 1
                    else:
                        walk_index -= 1

            self.selection_mode_range = False

            if thumbnail._selected:
                self.flowbox.unselect_child(thumbnail)
                self.selection_mode_last_thumbnail_index = None
                thumbnail._selected = False
            else:
                self.flowbox.select_child(thumbnail)
                self.selection_mode_last_thumbnail_index = thumbnail.get_index()
                thumbnail._selected = True

            if len(self.flowbox.get_selected_children()) == 0:
                self.leave_selection_mode()
        else:
            self.window.card.init(thumbnail.manga)

    def on_manga_thumbnail_right_click(self, _gesture, _n_press, x, y):
        """Allow to enter in selection mode with a right click on a thumbnail"""
        if self.selection_mode:
            return Gdk.EVENT_PROPAGATE

        thumbnail = self.flowbox.get_child_at_pos(x, y)
        if thumbnail is not None:
            self.enter_selection_mode()
            self.on_manga_thumbnail_activated(None, thumbnail)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def on_manga_updated(self, _obj, manga, _chapters_changes, _synced):
        self.refresh_on_manga_state_changed(manga)

    def on_overlaysplitview_revealed(self, _overlaysplitview, _param):
        if self.overlaysplitview.get_show_sidebar():
            self.categories_list.populate()
        else:
            self.categories_list.set_edit_mode(False)

    def on_resize(self):
        if self.page == 'start_page':
            return

        def do_resize():
            self.compute_thumbnails_cover_size()

            for thumbnail in self.flowbox:
                thumbnail.resize(*self.thumbnails_cover_size)

        # Wait until there are no higher priority events pending to the default main loop
        GLib.idle_add(do_resize)

    def on_search_entry_activated(self, _entry):
        """Open first manga in search when <Enter> is pressed"""
        thumbnail = self.flowbox.get_child_at_pos(0, 0)
        if thumbnail:
            self.on_manga_thumbnail_activated(None, thumbnail)

    def on_shown(self, _page):
        if self.searchbar.get_search_mode():
            self.search_entry.grab_focus()

        self.show_activity_indicator()

    def on_sort_order_changed(self, _action, variant):
        value = variant.get_string()
        if value == Settings.get_default().library_sort_order:
            return

        Settings.get_default().library_sort_order = value
        self.set_sort_order()

    def open_categories_editor(self, _action, _gparam):
        self.window.categories_editor.show()

    def open_download_manager(self, _action, _gparam):
        self.window.download_manager.show()

    def open_history(self, _action, _gparam):
        self.window.history.show()

    def populate(self):
        if self.populating:
            return True

        self.populating = True

        self.show_page('start_page')

        db_conn = create_db_connection()

        self.update_title(db_conn=db_conn)

        mangas_rows = db_conn.execute('SELECT id FROM mangas WHERE in_library = 1 ORDER BY last_read DESC').fetchall()

        paintable = Gdk.Texture.new_from_resource('/info/febvre/Komikku/images/logo.png')
        self.start_page_logo_image.set_from_paintable(paintable)

        if len(mangas_rows) == 0:
            self.start_page_title_label.set_markup('<span weight="bold">' + _('Welcome to Komikku') + '</span>')
            self.start_page_progressbar.set_visible(False)
            self.start_page_discover_button.set_visible(True)
            self.filters_button.set_visible(False)

            self.populating = False

            return

        self.start_page_progressbar.set_visible(True)
        self.start_page_progressbar.set_fraction(0)

        self.start_page_title_label.set_markup('<span weight="bold">' + _('Loading…') + '</span>')
        self.start_page_discover_button.set_visible(False)

        # Clear library flowbox
        thumbnail = self.flowbox.get_first_child()
        while thumbnail:
            next_thumbnail = thumbnail.get_next_sibling()
            self.flowbox.remove(thumbnail)
            thumbnail = next_thumbnail

        def run():
            db_conn = create_db_connection()

            for index, row in enumerate(mangas_rows):
                thumbnail = Thumbnail(self, Manga.get(row['id'], db_conn=db_conn), *self.thumbnails_cover_size)
                GLib.idle_add(on_progress, thumbnail, (index + 1) / len(mangas_rows))

            GLib.idle_add(on_complete)

        def on_complete():
            self.filters_button.set_visible(True)
            self.show_page('flowbox')
            self.populating = False

        def on_progress(thumbnail, fraction):
            self.start_page_progressbar.set_fraction(fraction)
            self.flowbox.append(thumbnail)

        # Populate flowbox
        self.compute_thumbnails_cover_size()

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def refresh_on_manga_state_changed(self, manga):
        # Update thumbnail
        self.update_thumbnail(manga)

        # Update filtering if filters are selected
        if self.selected_filters:
            self.flowbox.invalidate_filter()

    def remove_thumbnail(self, manga):
        # Remove manga thumbnail in flowbox
        thumbnail = self.flowbox.get_first_child()
        while thumbnail:
            if thumbnail.manga.id == manga.id:
                self.flowbox.remove(thumbnail)
                break
            thumbnail = thumbnail.get_next_sibling()

    def search(self, _search_entry):
        self.flowbox.invalidate_filter()

    def select_all(self, _action=None, _param=None):
        if self.page != 'flowbox':
            return

        if not self.selection_mode:
            self.enter_selection_mode()

        for thumbnail in self.flowbox:
            if not thumbnail._selected and not thumbnail._filtered:
                thumbnail._selected = True
                self.flowbox.select_child(thumbnail)

    def set_sort_order(self, invalidate=True):
        self.sort_order_action.set_state(GLib.Variant('s', Settings.get_default().library_sort_order))
        if invalidate:
            self.flowbox.invalidate_sort()

    def show_activity_indicator(self):
        def pulse():
            if self.window.page != self.props.tag or self.window.updater.update_library_flag:
                # Stop indicator
                self.activity_progressbar.props.fraction = 0
                return GLib.SOURCE_REMOVE

            if self.window.updater.running:
                self.activity_progressbar.pulse()
                return GLib.SOURCE_CONTINUE

            # Update is ended, stop indicator
            self.activity_progressbar.props.fraction = 0
            return GLib.SOURCE_REMOVE

        GLib.timeout_add(250, pulse)

    def show_page(self, name):
        if self.page == name:
            return

        self.page = name

        self.stack.set_visible_child_name(name)
        self.update_headerbar_buttons()

    def toggle_selected_read_status(self, _action, _param, read):
        chapters_ids = []
        chapters_data = []

        self.window.activity_indicator.set_visible(True)

        for thumbnail in self.flowbox.get_selected_children():
            for chapter in thumbnail.manga.chapters:
                chapters_ids.append(chapter.id)
                chapters_data.append(dict(
                    last_page_read_index=None,
                    read_progress=None,
                    read=read,
                    recent=False,
                ))

        db_conn = create_db_connection()
        with db_conn:
            res = update_rows(db_conn, 'chapters', chapters_ids, chapters_data)

        self.window.activity_indicator.set_visible(False)

        if not res:
            self.window.add_notification(_('Failed to update reading status'))
        else:
            if Settings.get_default().library_badges:
                for thumbnail in self.flowbox.get_selected_children():
                    thumbnail.update()

            self.leave_selection_mode()

            if self.selected_filters:
                self.flowbox.invalidate_filter()

    def update_all(self, _action, _param):
        self.window.updater.update_library()

    def update_headerbar_buttons(self):
        if self.page == 'flowbox':
            if self.selection_mode:
                self.left_button.set_label(_('Cancel'))
                self.left_button.set_tooltip_text(_('Cancel'))
                self.search_button.set_visible(False)
                self.menu_button.set_visible(False)
            else:
                self.left_button.set_tooltip_text(_('Add New Comic'))
                self.left_button.set_icon_name('list-add-symbolic')
                self.search_button.set_visible(True)
                self.menu_button.set_visible(True)

            self.categories_togglebutton.set_visible(True)
        else:
            self.categories_togglebutton.set_visible(False)
            self.search_button.set_visible(False)
            self.menu_button.set_visible(True)

    def update_selected(self, _action, _param):
        self.window.updater.add([thumbnail.manga for thumbnail in self.flowbox.get_selected_children()])
        if self.window.updater.start():
            self.show_activity_indicator()

        self.leave_selection_mode()

    def update_thumbnail(self, manga):
        for thumbnail in self.flowbox:
            if thumbnail.manga.id == manga.id:
                thumbnail.update(manga)
                break

    def update_title(self, *args, db_conn=None):
        title = 'Komikku'

        nb_selected = len(self.flowbox.get_selected_children()) if self.selection_mode else 0
        if nb_selected > 0:
            title = ngettext('{0} selected', '{0} selected', nb_selected).format(nb_selected)
        else:
            if (category_id := Settings.get_default().selected_category) != CategoryVirtual.ALL:
                if category_id == CategoryVirtual.UNCATEGORIZED:
                    title = _('Uncategorized')
                elif category := Category.get(category_id, db_conn=db_conn):
                    title = category.label
                else:
                    # Category saved in settings no longer exists!
                    # This should never happen, but if it does, reset selected category
                    Settings.get_default().selected_category = CategoryVirtual.ALL

        self.title.set_title(title)



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
import gc
import importlib
import json
import logging
import os
from pathlib import Path
import shutil
import time

from colorthief import ColorThief
from PIL import Image
from PIL import ImageFilter
from PIL import ImageStat

from komikku.consts import COVER_HEIGHT
from komikku.consts import COVER_WIDTH
from komikku.models.database import create_db_connection
from komikku.models.database import insert_row
from komikku.models.database import update_row
from komikku.models.database import update_rows
from komikku.servers.utils import get_server_class_name_by_id
from komikku.servers.utils import get_server_dir_name_by_id
from komikku.servers.utils import get_server_module_name_by_id
from komikku.utils import get_cached_data_dir
from komikku.utils import get_data_dir
from komikku.utils import is_number
from komikku.utils import markdown_to_markup
from komikku.utils import remove_number_leading_zero
from komikku.utils import trunc_filename

logger = logging.getLogger(__name__)


class Manga:
    _chapters = None
    _chapters_scanlators = None
    _server = None

    STATUSES = dict(
        complete=_('Complete'),
        ongoing=_('Ongoing'),
        suspended=_('Suspended'),
        hiatus=_('Hiatus'),
    )

    def __init__(self, server=None):
        if server:
            self._server = server

    @classmethod
    def get(cls, id_, server=None, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        row = db_conn.execute('SELECT * FROM mangas WHERE id = ?', (id_,)).fetchone()

        if row is None:
            return None

        manga = cls(server=server)
        for key in row.keys():
            setattr(manga, key, row[key])

        return manga

    @classmethod
    def new(cls, data, server, long_strip_detection):
        data = data.copy()
        chapters = data.pop('chapters')
        cover_url = data.pop('cover')

        # Remove search-specific data not saved in database
        for key in ('nb_chapters', 'nb_volumes', 'last_chapter', 'last_volume'):
            data.pop(key, None)
            continue

        if data.get('synopsis'):
            data['synopsis'] = markdown_to_markup(data['synopsis'])

        # Fill data with internal data
        data.update(dict(
            in_library=0,
            # Add fake last_read date: allows to display recently added manga at the top of the library
            last_read=datetime.datetime.now(datetime.UTC),
        ))

        # Long strip detection (Webtoon)
        if long_strip_detection and server.is_long_strip(data):
            data.update(dict(
                reading_mode='webtoon',
                scaling='width',
            ))

        db_conn = create_db_connection()
        with db_conn:
            id_ = insert_row(db_conn, 'mangas', data)

            rank = 0
            for chapter_data in chapters:
                if not chapter_data.get('date'):
                    # Used today if not date is provided
                    chapter_data['date'] = datetime.date.today()

                chapter = Chapter.new(chapter_data, rank, id_, db_conn)
                if chapter is not None:
                    rank += 1

        manga = cls.get(id_, server)

        if not os.path.exists(manga.path):
            os.makedirs(manga.path)

        manga._save_cover(cover_url)

        return manga

    @property
    def backdrop_colors_css(self):
        cover_path = self.cover_fs_path
        if cover_path is None:
            return None

        path = os.path.join(self.path, 'backdrop_colors.css')
        if os.path.exists(path):
            with open(path) as fp:
                data = fp.read()
                # CSS must be regenerated if old format is detected
                # Named colors are deprecated and will be removed in GTK5
                if '@define-color' not in data:
                    return data

        palette = ColorThief(cover_path).get_palette(color_count=2, quality=1)[:2]
        if len(palette) != 2:
            # Single color image?
            return None

        colors = [':root {\n']
        for index, color in enumerate(reversed(palette)):
            colors.append(f'\t--backdrop-background-color-{index}: rgb({color[0]} {color[1]} {color[2]} / 100%);\n')  # noqa: E702, E231
        colors.append('\t--backdrop-background-color-2: var(--window-bg-color);\n')
        colors.append('}\n')

        with open(path, 'w') as fp:
            fp.writelines(colors)

        return ''.join(colors)

    @property
    def backdrop_image_fs_path(self):
        if self.cover_fs_path is None:
            return None

        path = os.path.join(self.path, 'backdrop_image.jpg')
        if os.path.exists(path):
            return path

        with Image.open(self.cover_fs_path) as image:
            radius = 25 * max(1, image.width / 180)
            image = image.convert('RGB').filter(ImageFilter.GaussianBlur(radius))
            image.save(path, 'JPEG')

        return path

    @property
    def backdrop_info(self):
        if self.backdrop_image_fs_path is None:
            return None

        path = os.path.join(self.path, 'backdrop_info.json')
        if os.path.exists(path):
            with open(path) as fp:
                return json.load(fp)

        with Image.open(self.backdrop_image_fs_path) as image:
            stat = ImageStat.Stat(image.convert('L'))
            info = {
                # Luninance values used to apply an opacity on Picture depending of color scheme (dark/light)
                'luminance': [
                    min((stat.mean[0] + stat.extrema[0][0]) / 510, 0.7),
                    max((stat.mean[0] + stat.extrema[0][1]) / 510, 0.3),
                ]
            }
            with open(path, 'w') as fp:
                json.dump(info, fp)

        return info

    @property
    def categories(self):
        db_conn = create_db_connection()
        rows = db_conn.execute(
            'SELECT c.id FROM categories c JOIN categories_mangas_association cma ON cma.category_id = c.id WHERE cma.manga_id = ?',
            (self.id,)
        )

        categories = []
        for row in rows:
            categories.append(row['id'])

        return categories

    @property
    def chapters(self):
        if self._chapters is None:
            db_conn = create_db_connection()
            if self.sort_order and self.sort_order.endswith('asc'):
                rows = db_conn.execute('SELECT * FROM chapters WHERE manga_id = ? ORDER BY rank ASC', (self.id,))
            else:
                rows = db_conn.execute('SELECT * FROM chapters WHERE manga_id = ? ORDER BY rank DESC', (self.id,))

            self._chapters = []
            for row in rows:
                self._chapters.append(Chapter(row=row, manga=self))

        return self._chapters

    @property
    def chapters_scanlators(self):
        if self._chapters_scanlators is None:
            db_conn = create_db_connection()

            rows = db_conn.execute('SELECT DISTINCT scanlators, count(*) FROM chapters WHERE manga_id = ? GROUP BY scanlators', (self.id,))

            scanlators = {}
            for row in rows:
                if not row[0]:  # None or []
                    # Use 'Unknown' as virtual scanlator for chapters without scanlators defined
                    row = (['Unknown'], row[1])

                for scanlator in row[0]:
                    if scanlator not in scanlators:
                        scanlators[scanlator] = {
                            'name': scanlator,
                            'count': 0,
                        }

                    scanlators[scanlator]['count'] += row[1]

            self._chapters_scanlators = list(scanlators.values()) or None

        return self._chapters_scanlators

    @property
    def class_name(self):
        return get_server_class_name_by_id(self.server_id)

    @property
    def cover_fs_path(self):
        path = os.path.join(self.path, 'cover.jpg')
        if os.path.exists(path):
            return path

        return None

    @property
    def dir_name(self):
        return get_server_dir_name_by_id(self.server_id)

    @property
    def is_local(self):
        return self.server_id == 'local'

    @property
    def module_name(self):
        return get_server_module_name_by_id(self.server_id)

    @property
    def nb_downloaded_chapters(self):
        db_conn = create_db_connection()
        row = db_conn.execute(
            'SELECT count() AS downloaded FROM chapters WHERE manga_id = ? AND downloaded = 1 and read = 0', (self.id,)).fetchone()

        return row['downloaded']

    @property
    def nb_recent_chapters(self):
        db_conn = create_db_connection()
        row = db_conn.execute('SELECT count() AS recents FROM chapters WHERE manga_id = ? AND recent = 1', (self.id,)).fetchone()

        return row['recents']

    @property
    def nb_unread_chapters(self):
        db_conn = create_db_connection()
        row = db_conn.execute('SELECT count() AS unread FROM chapters WHERE manga_id = ? AND read = 0', (self.id,)).fetchone()

        return row['unread']

    @property
    def last_read_chapter(self):
        db_conn = create_db_connection()
        row = db_conn.execute(
            'SELECT * FROM chapters WHERE manga_id = ? AND last_read IS NOT NULL ORDER BY last_read DESC LIMIT 1', (self.id,)
        ).fetchone()

        return Chapter(row=row, manga=self) if row else None

    @property
    def path(self):
        if self.in_library:
            return os.path.join(get_data_dir(), self.dir_name, trunc_filename(self.name))

        return os.path.join(get_cached_data_dir(), self.dir_name, trunc_filename(self.name))

    @property
    def server(self):
        if self._server is None:
            try:
                module = importlib.import_module('.' + self.module_name, package='komikku.servers')
                self._server = getattr(module, self.class_name)()
            except Exception:
                from komikku.servers import ServerDummy

                self._server = ServerDummy(self.server_id)
                logger.info('Failed to import server: %s', self.class_name)

        return self._server

    def _save_cover(self, url):
        # Covers in landscape format are converted to portrait format
        if self.server.save_image(url, self.path, 'cover', COVER_WIDTH, COVER_HEIGHT):
            # Remove backdrop files (image, css, info)
            for file in Path(self.path).glob('backdrop_*'):
                os.unlink(file)

    def add_in_library(self):
        tmp_path = self.path

        self.update(dict(in_library=True))

        if self.is_local:
            # Move files
            for filename in os.listdir(tmp_path):
                dst_path = os.path.join(self.path, filename)
                if os.path.exists(dst_path):
                    os.remove(dst_path)
                shutil.move(os.path.join(tmp_path, filename), self.path)

            # Remove folder
            shutil.rmtree(tmp_path)
        else:
            # Move folder
            shutil.move(tmp_path, self.path)

    def clear(self):
        # Remove chapters folders
        for path in Path(self.path).glob('*'):
            if path.is_dir():
                shutil.rmtree(path)

        # Set all chapters downloaded flag to 0
        db_conn = create_db_connection()
        with db_conn:
            db_conn.execute('UPDATE chapters SET downloaded = 0 WHERE manga_id = ?', (self.id, ))

        self._chapters = None

    def delete(self, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            db_conn.execute('DELETE FROM mangas WHERE id = ?', (self.id, ))

        # Delete folder except when server is 'local'
        if os.path.exists(self.path) and not self.is_local:
            shutil.rmtree(self.path)

    def get_next_chapter(self, chapter, direction=1):
        """
        :param chapter: reference chapter
        :param direction: -1 for preceding chapter, 1 for following chapter
        """
        assert direction in (-1, 1), 'Invalid direction value'

        db_conn = create_db_connection()

        op = '>' if direction == 1 else '<'
        order = 'ASC' if direction == 1 else 'DESC'
        if self.filters and self.filters.get('scanlators'):
            # Chapters must be filtered by scanlators (some scanlators are excluded)
            excluded_scanlators = self.filters['scanlators']

            # Subquery to get IDs of not filtered chapters
            scanlators_subquery = f"""
                SELECT DISTINCT c.id
                FROM chapters c, json_each(scanlators)
                WHERE json_each.value NOT IN ("{'", "'.join(excluded_scanlators)}") AND c.manga_id = {self.id}
            """
            if 'Unknown' not in excluded_scanlators:
                # Add chapters without scanlators defined
                scanlators_subquery += f"""
                    UNION
                    SELECT id FROM chapters WHERE (scanlators IS NULL OR scanlators->0 IS NULL) AND manga_id = {self.id}
                """
        else:
            scanlators_subquery = None

        if self.sort_order in ('asc', 'desc', None):
            if scanlators_subquery:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND id IN ({scanlators_subquery}) AND rank {op} ? ORDER BY rank {order}',
                    (self.id, chapter.rank)
                ).fetchone()
            else:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND rank {op} ? ORDER BY rank {order}',
                    (self.id, chapter.rank)
                ).fetchone()

        elif self.sort_order in ('date-asc', 'date-desc'):
            if scanlators_subquery:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND id IN ({scanlators_subquery}) AND date {op} ? ORDER BY date {order}, id {order}',
                    (self.id, chapter.date)
                ).fetchone()
            else:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND date {op} ? ORDER BY date {order}, id {order}',
                    (self.id, chapter.date)
                ).fetchone()

        elif self.sort_order in ('natural-asc', 'natural-desc'):
            if scanlators_subquery:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND id IN ({scanlators_subquery}) AND title {op} ? COLLATE natsort ORDER BY title {order}, id {order}',
                    (self.id, chapter.title)
                ).fetchone()
            else:
                row = db_conn.execute(
                    f'SELECT * FROM chapters WHERE manga_id = ? AND title {op} ? COLLATE natsort ORDER BY title {order}, id {order}',
                    (self.id, chapter.title)
                ).fetchone()

        if not row:
            return None

        return Chapter(row=row, manga=self)

    def toggle_category(self, category_id, active):
        db_conn = create_db_connection()
        with db_conn:
            if active:
                insert_row(db_conn, 'categories_mangas_association', dict(category_id=category_id, manga_id=self.id))
            else:
                db_conn.execute(
                    'DELETE FROM categories_mangas_association WHERE category_id = ? AND manga_id = ?',
                    (category_id, self.id,)
                )

    def update(self, data, db_conn=None):
        """
        Updates specific fields

        :param dict data: fields to update
        :return: True on success False otherwise
        """
        ret = False

        # Update
        for key in data:
            setattr(self, key, data[key])

        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            ret = update_row(db_conn, 'mangas', self.id, data)

        return ret

    def update_full(self, db_conn=None):
        """
        Updates manga

        :return: True on success False otherwise, chapters_changes, synced
        :rtype: tuple
        """
        chapters_changes = {
            'recent_ids': [],
            'nb_updated': 0,
            'nb_deleted': 0,
        }
        gone_chapters_ranks = []

        def get_free_rank(rank):
            if rank not in gone_chapters_ranks:
                return rank

            return get_free_rank(rank + 1)

        data = self.server.get_manga_data(dict(
            slug=self.slug,
            name=self.name,
            url=self.url,
            last_read=self.last_read
        ))
        gc.collect()

        if data is None:
            return False, chapters_changes, False

        synced = self.server.sync and data['last_read'] != self.last_read

        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            # Update chapters
            chapters_data = data.pop('chapters')

            # First, delete chapters that no longer exist on server EXCEPT those marked as downloaded
            # If server is 'local', chapters are always deleted
            chapters_slugs = [str(chapter_data['slug']) for chapter_data in chapters_data]
            rows = db_conn.execute('SELECT * FROM chapters WHERE manga_id = ?', (self.id,))
            for row in rows:
                if row['slug'] not in chapters_slugs:
                    gone_chapter = Chapter.get(row['id'], manga=self, db_conn=db_conn)
                    if not gone_chapter.downloaded or self.is_local:
                        # Chapter is not downloaded or server is 'local'
                        # Delete chapter
                        gone_chapter.delete(db_conn)
                        chapters_changes['nb_deleted'] += 1

                        logger.warning(
                            '[UPDATE] {0} ({1}): Delete chapter {2} (no longer available)'.format(
                                self.name, self.server_id, gone_chapter.title
                            )
                        )
                    else:
                        # Chapter is downloaded
                        # Keep track of rank because it must not be reused
                        gone_chapters_ranks.append(gone_chapter.rank)

            # Then, add or update chapters
            rank = 0
            for chapter_data in chapters_data:
                row = db_conn.execute(
                    'SELECT * FROM chapters WHERE manga_id = ? AND slug = ?', (self.id, chapter_data['slug'])
                ).fetchone()

                rank = get_free_rank(rank)
                if row:
                    # Update chapter
                    changes = {}

                    # Common fields
                    for key in ('title', 'num', 'num_volume', 'url', 'date', 'scanlators'):
                        if row[key] != chapter_data.get(key):
                            if key in ('num', 'num_volume'):
                                num = str(chapter_data.get(key))
                                changes[key] = remove_number_leading_zero(num) if is_number(num) else None
                            else:
                                changes[key] = chapter_data.get(key)

                    if row['rank'] != rank:
                        changes['rank'] = rank

                    if changes:
                        chapters_changes['nb_updated'] += 1

                    # Sync fields
                    for key in ('last_page_read_index', 'last_read', 'read'):
                        if chapter_data.get(key) and row[key] != chapter_data[key]:
                            changes[key] = chapter_data[key]

                    if changes:
                        update_row(db_conn, 'chapters', row['id'], changes)

                    rank += 1
                else:
                    # Add new chapter

                    # Ensure chapter num and volume num are numbers
                    for key in ('num', 'num_volume'):
                        if chapter_data.get(key) is not None:
                            num = str(chapter_data[key])
                            chapter_data[key] = remove_number_leading_zero(num) if is_number(num) else None

                    # Used today if not date is provided
                    if not chapter_data.get('date'):
                        chapter_data['date'] = datetime.date.today()

                    chapter_data.update(dict(
                        manga_id=self.id,
                        rank=rank,
                        downloaded=chapter_data.get('downloaded', 0),
                        recent=1,
                        read=0,
                    ))
                    id_ = insert_row(db_conn, 'chapters', chapter_data)
                    if id_ is not None:
                        chapters_changes['recent_ids'].append(id_)
                        rank += 1

                        logger.info('[UPDATE] {0} ({1}): Add new chapter {2}'.format(self.name, self.server_id, chapter_data['title']))

            if not data.get('last_update'):
                if chapters_changes['recent_ids'] or chapters_changes['nb_updated'] or chapters_changes['nb_deleted']:
                    data['last_update'] = datetime.datetime.now(datetime.UTC)

            # Update cover
            cover = data.pop('cover', None)
            if cover:
                self._save_cover(cover)

            # Convert synopsis Markdown links (if any) into HTML <a> tags
            if data.get('synopsis'):
                data['synopsis'] = markdown_to_markup(data['synopsis'])

            # Clear cache
            self._chapters = None
            self._chapters_scanlators = None

            # Store old path
            old_path = self.path

            # Update
            for key in data:
                setattr(self, key, data[key])

            update_row(db_conn, 'mangas', self.id, data)

            if old_path != self.path:
                # Manga name changes, manga folder must be renamed too
                os.rename(old_path, self.path)

        return True, chapters_changes, synced


class Chapter:
    _manga = None

    def __init__(self, row=None, manga=None):
        if row is not None:
            if manga:
                self._manga = manga
            for key in row.keys():
                setattr(self, key, row[key])

    @classmethod
    def get(cls, id_, manga=None, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        row = db_conn.execute('SELECT * FROM chapters WHERE id = ?', (id_,)).fetchone()

        if row is None:
            return None

        return cls(row, manga)

    @classmethod
    def new(cls, data, rank, manga_id, db_conn=None):
        # Fill data with internal data
        data = data.copy()
        data.update(dict(
            manga_id=manga_id,
            rank=rank,
            downloaded=data.get('downloaded', 0),
            recent=0,
            read=0,
        ))

        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            id_ = insert_row(db_conn, 'chapters', data)

        chapter = cls.get(id_, db_conn=db_conn) if id_ is not None else None

        return chapter

    @property
    def clearable(self):
        # Not clearable if server is 'local'
        return os.path.exists(self.path) and not self.is_local

    @property
    def is_local(self):
        return self.manga.is_local

    @property
    def manga(self):
        if self._manga is None:
            self._manga = Manga.get(self.manga_id)

        return self._manga

    @property
    def number(self):
        """ Returns chapter number"""
        if self.num and is_number(self.num):
            return remove_number_leading_zero(self.num)

        logger.warning(f'{self.manga.name} serie ({self.manga.server_id}) do not support tracking (no chapter num)?')

        return None

    @property
    def path(self):
        # BEWARE: self.slug may contain '/' characters
        # os.makedirs() must be used to create chapter's folder
        name = '/'.join([trunc_filename(part) for part in self.slug.split('/')])

        return os.path.join(self.manga.path, name)

    def clear(self, reset=False):
        """
        Clear (erase files on disk) and optionally reset

        :param bool reset: reset
        :return: True on success False otherwise
        """
        if self.clearable:
            shutil.rmtree(self.path)

        data = dict(
            downloaded=0,
        )
        if reset:
            data.update(dict(
                pages=None,
                read_progress=None,
                read=0,
                last_read=None,
                last_page_read_index=None,
            ))

        return self.update(data)

    @staticmethod
    def clear_many(chapters, reset=False):
        # Assume all chapters belong to the same manga
        manga = chapters[0].manga
        ids = []
        data = []

        for chapter in chapters:
            # Delete folder except when server is 'local'
            if os.path.exists(chapter.path) and not manga.is_local:
                shutil.rmtree(chapter.path)

            ids.append(chapter.id)

            updated_data = dict(
                downloaded=0,
            )
            if reset:
                updated_data.update(dict(
                    pages=None,
                    read_progress=None,
                    read=0,
                    last_read=None,
                    last_page_read_index=None,
                ))
            data.append(updated_data)

        db_conn = create_db_connection()
        with db_conn:
            update_rows(db_conn, 'chapters', ids, data)

    def delete(self, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            db_conn.execute('DELETE FROM chapters WHERE id = ?', (self.id, ))

        if os.path.exists(self.path):
            shutil.rmtree(self.path)

    def get_page(self, index, db_conn=None):
        page_path = self.get_page_path(index)
        if page_path:
            return page_path, None

        page = self.pages[index]

        start = time.perf_counter()
        data = self.manga.server.get_manga_chapter_page_image(self.manga.slug, self.manga.name, self.slug, page)
        rtime = time.perf_counter() - start
        gc.collect()

        if data is None:
            return None, None

        if not os.path.exists(self.path):
            os.makedirs(self.path, exist_ok=True)

        image = data['buffer']
        page_path = os.path.join(self.path, data['name'])
        with open(page_path, 'wb') as fp:
            fp.write(image)

        updated_data = {}

        # If page name can't be retrieved from `image` or `slug`, we store its name
        retrievable = False
        if page.get('image') and data['name'] == page['image'].split('?')[0].split('/')[-1]:
            retrievable = True
        elif page.get('slug') and data['name'] == page['slug'].split('/')[-1]:
            retrievable = True
        if not retrievable:
            self.pages[index]['name'] = data['name']
            updated_data['pages'] = self.pages

        downloaded = len(next(os.walk(self.path))[2]) == len(self.pages)
        if downloaded != self.downloaded:
            updated_data['downloaded'] = downloaded

        if updated_data:
            self.update(updated_data, db_conn=db_conn)

        return page_path, rtime

    def get_page_data(self, index):
        """
        Return page image data: buffer, MIME type, name

        Useful for locally stored manga. Image data (bytes) are retrieved directly from archive.
        """
        return self.manga.server.get_manga_chapter_page_image(self.manga.slug, self.manga.name, self.slug, self.pages[index])

    def get_page_path(self, index):
        if not self.pages:
            return None

        page = self.pages[index]

        # Get image name
        if page.get('name'):
            name = page['name']

        elif page.get('image') and page['image'].split('/')[-1]:
            # Extract from URL (relative or absolute)
            name = page['image'].split('/')[-1]
            # Remove query string
            name = name.split('?')[0]

        elif page.get('slug') and page['slug'].split('/')[-1]:
            name = page['slug'].split('/')[-1]

        else:
            return None

        path = os.path.join(self.path, name)

        return path if os.path.exists(path) else None

    def update(self, data, db_conn=None):
        """
        Updates specific fields

        :param dict data: fields to update
        :return: True on success False otherwise
        """
        ret = False

        for key in data:
            setattr(self, key, data[key])

        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            ret = update_row(db_conn, 'chapters', self.id, data)

        return ret

    def update_full(self, db_conn=None):
        """
        Updates chapter

        Fetches server and saves chapter data

        :return: True on success False otherwise
        """
        if self.pages:
            return True

        data = self.manga.server.get_manga_chapter_data(self.manga.slug, self.manga.name, self.slug, self.url)
        gc.collect()

        if data is None or not data['pages']:
            return False

        return self.update(data, db_conn=db_conn)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.wpcomics import WPComics
from komikku.servers.utils import convert_date_string
from komikku.utils import is_number


class Doctruyen3q(WPComics):
    id = 'doctruyen3q'
    name = 'DocTruyen3Q'
    lang = 'vi'
    is_nsfw = True

    has_cf = True

    base_url = 'https://doctruyen3qhubx.com'
    search_url = base_url + '/tim-truyen'
    latest_updates_url = base_url + '/tim-truyen?sort=1'
    most_populars_url = base_url + '/tim-truyen?sort=2'
    manga_url = base_url + '/truyen-tranh/{0}'
    chapter_url = base_url + '/truyen-tranh/{0}/{1}'

    chapters_name = 'chapter'
    date_format = None
    image_src_attrs = ['data-original', 'data-src', 'src']
    slug_segments = 2

    details_name_selector = 'h1.title-manga'
    details_cover_selector = 'img.image-comic'
    details_status_selector = None
    details_authors_selector = None
    details_genres_selector = '.category a'
    details_synopsis_selector = '.detail-summary'

    results_selector = '.item-manga'
    results_latest_selector = '.item-manga'
    results_link_selector = '.image-item a'
    results_cover_img_selector = '.image-item a img'
    results_last_chapter_link_selector = '.caption ul li a'
    results_last_chapter_lastest_updates_link_selector = '.caption ul li a'

    chapters_selector = '#list-chapter-dt ul li.row'

    long_strip_genres = [
        'Manhua',
        'Manhwa',
    ]

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        manga_slug = manga_slug.split('/')[0]

        return super(Doctruyen3q, self).get_manga_chapter_data(manga_slug, manga_name, chapter_slug, chapter_url)

    def get_manga_chapters_data(self, slug, soup):
        """
        Returns manga chapters list
        """
        chapters = []
        for element in soup.select(self.chapters_selector):
            a_element = element.select_one('div a')
            url = a_element.get('href')
            slug = url.split('/')[-2]
            num = slug.split('-')[-1] if slug.startswith(f'{self.chapters_name}-') else None
            date = element.select_one('div:last-child').text.strip()

            chapters.append({
                'slug': url.split('/', len(url.split('/')) - self.slug_segments)[-1],
                'title': a_element.text.strip(),
                'num': num if is_number(num) else None,
                'date': convert_date_string(date),
            })

        return list(reversed(chapters))



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import requests

from komikku.consts import USER_AGENT
from komikku.trackers import authenticated
from komikku.trackers import Tracker

# https://api.mangaupdates.com/


class Mangaupdates(Tracker):
    id = 'mangaupdates'
    name = 'MangaUpdates'

    base_url = 'https://www.mangaupdates.com'
    api_url = 'https://api.mangaupdates.com/v1'
    api_login_url = api_url + '/account/loginWithCookie'
    api_refresh_token_url = api_url + '/account/refresh'
    api_search_url = api_url + '/series/search'
    api_manga_url = api_url + '/series/{0}'
    api_list_add_url = api_url + '/lists/series'
    api_list_update_url = api_url + '/lists/series/update'
    api_manga_list_url = api_url + '/lists/series/{0}'
    api_manga_rating_url = api_url + '/series/{0}/rating'

    STATUSES_MAPPING = {
        # tracker status (list) => internal status
        'read': 'reading',
        'complete': 'completed',
        'hold': 'on_hold',
        'unfinished': 'dropped',
        'wish': 'plan_to_read',
    }

    STATUSES_ID_MAPPING = {
        # internal status => tracker status id (list id)
        'reading': 0,
        'completed': 2,
        'on_hold': 4,
        'dropped': 3,
        'plan_to_read': 1,
    }

    USER_SCORE_FORMAT = {
        'min': 0,
        'max': 10,
        'step': .1,
        'raw_factor': 1,
    }

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_url(self, id):
        return None

    def get_tracker_manga_data(self, id):
        # Get serie info
        r = self.session.get(self.api_manga_url.format(id))
        if r.status_code != 200:
            return None

        data = r.json()

        return {
            'id': data['series_id'],
            'name': data['title'],
            'chapters': data['latest_chapter'],
            'score_format': None,
            'url': data['url'],
        }

    def get_user_score_format(self, format):
        return self.USER_SCORE_FORMAT

    @authenticated
    def get_user_manga_data(self, id, access_token=None):
        # Get user serie info: status, chapters progress
        r = self.session_get(
            self.api_manga_list_url.format(id),
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        # Get serie info
        data = self.get_tracker_manga_data(id)
        if data is None:
            return None

        serie_status = None
        for status, list_id in self.STATUSES_ID_MAPPING.items():
            if list_id == resp_data['list_id']:
                serie_status = status
                break

        if serie_status is None:
            return None

        data.update({
            'chapters_progress': resp_data['status']['chapter'],
            'status': serie_status,
        })

        # Get user serie score
        r = self.session_get(
            self.api_manga_rating_url.format(id),
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        data['score'] = r.json()['rating'] if r.status_code == 200 else 0

        return data

    def refresh_access_token(self):
        """Returns new access token"""
        r = self.session_get(
            self.api_refresh_token_url,
            headers={
                'Authorization': f'Bearer {self.data["access_token"]}',
            }
        )
        if r.status_code != 200:
            return None

        data = r.json()
        if data['status'] != 'success':
            return None

        for cookie in self.session.cookies:
            if cookie.name != 'session':
                continue

            self.data = {
                'access_token': cookie.value,
                'refresh_token': None,
                'exp': cookie.expires,
            }
            return cookie.value

        return None

    def request_access_token(self, username, password):
        r = self.session_put(
            self.api_login_url,
            data={
                'username': username,
                'password': password,
            }
        )
        if r.status_code != 200:
            return False, 'failed'

        data = r.json()
        if data['status'] != 'success':
            return False, 'failed'

        for cookie in self.session.cookies:
            if cookie.name != 'session':
                continue

            self.data = {
                'access_token': cookie.value,
                'refresh_token': None,
                'exp': cookie.expires,
            }

            return True, None

        return False, 'failed'

    def search(self, term):
        r = self.session.post(
            self.api_search_url,
            data={
                'search': term,
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['results']:
            record = item['record']
            results.append({
                'id': record['series_id'],
                'cover': record['image']['url']['thumb'],
                'name': record['title'],
                'score': record['bayesian_rating'],
                'start_date': record['year'],
                'synopsis': record['description'],
            })

        return results

    @authenticated
    def update_user_manga_data(self, id, data, access_token=None):
        # Check if serie is already in a list
        r = self.session_get(
            self.api_manga_list_url.format(id),
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )

        if r.status_code == 200:
            # Update
            r = self.session_post(
                self.api_list_update_url,
                json=[{
                    'series': {
                        'id': id,
                    },
                    'list_id': self.STATUSES_ID_MAPPING[data['status']],
                    'status': {
                        'chapter': int(data['chapters_progress']),
                    }
                }],
                headers={
                    'Authorization': f'Bearer {access_token}',
                }
            )

        elif r.status_code == 404:
            # Add
            r = self.session_post(
                self.api_list_add_url,
                json=[
                    {
                        'series': {
                            'id': id,
                        },
                        'list_id': self.STATUSES_ID_MAPPING[data['status']],
                        'status': {
                            'volume': 1,
                            'chapter': int(data['chapters_progress']),
                            'increment_volume': 0,
                            'increment_chapter': 0,
                        },
                        'priority': 0,
                    }
                ],
                headers={
                    'Authorization': f'Bearer {access_token}',
                }
            )

        else:
            return False

        if r.status_code != 200:
            return False

        # Update score
        r = self.session_put(
            self.api_manga_rating_url.format(id),
            data={
                'rating': data['score'],
            },
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )

        if r.status_code != 200:
            return False

        return True



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from functools import wraps
import json
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

SERVER_NAME = 'MangaKawaii'


def set_lang(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        server = args[0]
        if not server.is_lang_set:
            server.session_get(
                server.base_url + '/lang/' + server.lang,
                params={
                    'Referer': f'{server.base_url}/',
                }
            )
            server.is_lang_set = True

        return func(*args, **kwargs)

    return wrapper


class Mangakawaii(Server):
    id = 'mangakawaii'
    name = SERVER_NAME
    lang = 'fr'
    has_cf = True
    is_nsfw = True
    long_strip_genres = ['Webtoon', ]

    base_url = 'https://www.mangakawaii.io'
    logo_url = base_url + '/assets/img/favicon-32x32.png'
    search_url = base_url + '/recherche-manga'
    manga_list_url = base_url + '/filterMangaList'
    manga_list_referer_url = base_url + '/liste-manga'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/manga/{0}/{1}/{2}/1'
    chapters_url = base_url + '/loadChapter?page={0}'
    cdn_base_urls = [
        'https://cdn.mangakawaii.io',
        'https://cdn2.mangakawaii.io',
    ]
    image_url = '{0}/uploads/manga/{1}/chapters_{2}/{3}/{4}?{5}'
    cover_url = cdn_base_urls[0] + '/uploads/manga/{0}/cover/cover_250x350.jpg'
    bypass_cf_url = base_url + '/manga/martial-peak'

    csrf_token = None
    is_lang_set = False

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            cookie = requests.cookies.create_cookie(
                name='mk_search_type',
                value='manga',
                domain=urlparse(self.base_url).netloc,
                path='/',
                expires=None,
            )
            self.session.cookies.set_cookie(cookie)
            cookie = requests.cookies.create_cookie(
                name='mk_cookie_consent',
                value='1',
                domain=urlparse(self.base_url).netloc,
                path='/',
                expires=None,
            )
            self.session.cookies.set_cookie(cookie)

            self.session.headers.update({'User-Agent': USER_AGENT})

    @CompleteChallenge()
    @set_lang
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        self.csrf_token = soup.select_one('meta[name="csrf-token"]')['content']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = soup.find('h1').text.strip()
        if data.get('cover') is None:
            data['cover'] = self.cover_url.format(data['slug'])

        # Details
        elements = soup.find('div', class_='col-md-8 mt-4 mt-md-0').find_all('dl')
        for element in elements:
            label = element.dt.text.strip()

            if label.startswith(('Author', 'Auteur', 'Artist', 'Artiste')):
                value = element.dd.span.text.strip()
                for t in value.split(','):
                    t = t.strip()
                    if t not in data['authors']:
                        data['authors'].append(t)

            elif label.startswith('Scantrad'):
                for a_element in element.dd.find_all('a', itemprop='name'):
                    data['scanlators'].append(a_element.text.replace('[', '').replace(']', '').strip())

            elif label.startswith('Genres'):
                a_elements = element.dd.find_all('a')
                data['genres'] = [a_element.text.strip() for a_element in a_elements]

            elif label.startswith(('Status', 'Statut')):
                status = element.dd.span.text.strip().lower()
                if status in ('ongoing', 'en cours'):
                    data['status'] = 'ongoing'
                elif status in ('completed', 'terminé'):
                    data['status'] = 'complete'
                elif status in ('abandoned', 'abandonné'):
                    data['status'] = 'suspended'
                elif status in ('paused', 'en pause'):
                    data['status'] = 'hiatus'

            elif label.startswith(('Summary', 'Description')):
                data['synopsis'] = element.dd.text.strip()

        #
        # Chapters
        # They are displayed in reverse order and loaded by page (if many)
        #
        # First, we get the oeuvreId
        oeuvre_id = None
        for script_element in reversed(soup.find_all('script')):
            script = script_element.string
            if not script or not script.strip().startswith('var ENDPOINT'):
                continue

            for line in script.split('\n'):
                line = line.strip()
                if not line.startswith('var oeuvreId'):
                    continue

                oeuvre_id = line.split("'")[-2]
                break

            break

        # Next, we retrieve first chapters available in page's HTML
        for tr_element in soup.find('table', class_='table--manga').find_all('tr'):
            td_element = tr_element.find('td', class_='table__chapter')
            if not td_element:
                continue

            slug = td_element.a.get('href').strip().split('/')[-1]
            date_element = tr_element.find('td', class_='table__date')
            if date_element.div:
                date_element.div.extract()

            data['chapters'].append(dict(
                slug=slug,
                title=' '.join(td_element.a.span.text.strip().split()),
                num=slug if is_number(slug) else None,
                date=convert_date_string(date_element.text.strip(), format='%d.%m.%Y'),
            ))

        # Finally, we recursively retrieve other chapters by page (via a web service)
        data['chapters'] += self.get_manga_chapters_data(data['slug'], oeuvre_id)
        data['chapters'].reverse()

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, self.lang, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for script_element in reversed(soup.find_all('script')):
            script = script_element.string
            if not script or not script.strip().startswith('var title'):
                continue

            for line in script.split('\n'):
                line = line.strip()
                if not line.startswith('var pages'):
                    continue

                pages = json.loads(line[12:-1])
                for index, page in enumerate(pages):
                    data['pages'].append(dict(
                        slug=None,
                        image=page['page_image'],
                        index=index,
                        version=page['page_version'],
                    ))

                break

        return data

    def get_manga_chapters_data(self, manga_slug, oeuvre_id, page=2):
        r = self.session_post(
            self.chapters_url.format(page),
            data=dict(
                oeuvreType='manga',
                oeuvreId=oeuvre_id,
                oeuvreSlug=manga_slug,
                oeuvreDownload='0',
            ),
            headers={
                'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                'Referer': self.manga_url.format(manga_slug),
                'X-CSRF-TOKEN': self.csrf_token,
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return []

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return []

        soup = BeautifulSoup(r.text, 'lxml')

        chapters = []
        for tr_element in soup.find_all('tr'):
            td_element = tr_element.find('td', class_='table__chapter')
            if not td_element:
                continue

            slug = td_element.a.get('href').strip().split('/')[-1]
            date_element = tr_element.find('td', class_='table__date')
            if date_element.div:
                date_element.div.extract()

            chapters.append(dict(
                slug=slug,
                title=' '.join(td_element.a.span.text.strip().split()),
                num=slug if is_number(slug) else None,
                date=convert_date_string(date_element.text.strip(), format='%d.%m.%Y'),
            ))

        chapters += self.get_manga_chapters_data(manga_slug, oeuvre_id, page + 1)

        return chapters

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        success = False
        for cdn_base_url in self.cdn_base_urls:
            r = self.session_get(
                self.image_url.format(cdn_base_url, manga_slug, self.lang, chapter_slug, page['image'], page['version']),
                headers={
                    'Referer': self.base_url,
                }
            )
            if r.ok:
                success = True
                break

        if not success:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, orderby):
        """
        Returns manga list sorted
        """
        params = dict(
            page=1,
            cat='',
            alpha='',
            asc='false',
            author='',
        )
        params['sortBy'] = 'views' if orderby == 'populars' else 'last_updated'

        r = self.session_get(
            self.manga_list_url,
            params=params,
            headers={
                'X-Requested-With': 'XMLHttpRequest',
                'Referer': self.manga_list_url,
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/plain':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.find_all('div', class_='media-thumbnail'):
            slug = element.find('a').get('href').split('/')[-1]
            results.append(dict(
                name=element.find('div', class_='media-thumbnail__overlay').find('h3').text.strip(),
                slug=slug,
                cover=self.cover_url.format(slug),
            ))

        return results

    @CompleteChallenge()
    @set_lang
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.get_manga_list('latest')

    @CompleteChallenge()
    @set_lang
    def get_most_populars(self):
        """
        Returns list of most viewed manga
        """
        return self.get_manga_list('populars')

    @CompleteChallenge()
    @set_lang
    def search(self, term):
        r = self.session_get(
            self.search_url,
            params=dict(query=term),
            headers={
                'X-Requested-With': 'XMLHttpRequest',
                'Referer': self.base_url,
            }
        )

        if r.status_code == 200:
            try:
                # Returned data for each manga:
                # value: name of the manga
                # slug: slug of the manga
                # imageUrl: cover of the manga
                data = r.json()['suggestions']

                results = []
                for item in data:
                    results.append(dict(
                        slug=item['slug'],
                        name=item['value'],
                        cover=self.cover_url.format(item['slug']),
                    ))
            except Exception:
                return None
            else:
                return results

        return None


class Mangakawaii_en(Mangakawaii):
    id = 'mangakawaii_en'
    lang = 'en'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# DocTruyen3Q [VI]
# JManga [JA] (disabled)
# Read Comics Free [EN] (disabled)
# Xoxocomics [EN]

import time
from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.utils import get_response_elapsed
from komikku.webview import CompleteChallenge

# WPComics Wordpress theme


class WPComics(Server):
    base_url: str = None
    search_url: str = None
    latest_updates_url: str = None
    most_populars_url: str = None
    manga_url: str = None
    chapter_url: str = None

    chapters_name: str = 'issue'
    date_format: str = '%m/%d/%Y'
    ignore_images_status_code: bool = False
    image_src_attrs: list = ['data-original']
    slug_segments: int = 1

    details_name_selector: str = None
    details_cover_selector: str = None
    details_status_selector: str = None
    details_authors_selector: str = None
    details_genres_selector: str = None
    details_synopsis_selector: str = None

    results_selector: str = '.items .item'
    results_latest_selector: str = '.list-chapter .row'
    results_link_selector: str = None
    results_cover_img_selector: str = None
    results_last_chapter_link_selector: str = None
    results_last_chapter_lastest_updates_link_selector: str = None

    chapters_selector: str = '#nt_listchapter ul li.row'

    def __init__(self):
        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers = {
                'User-Agent': USER_AGENT,
            }

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping manga HTML page content

        Initial data should contain at least comic's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug'], 1))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.select_one(self.details_name_selector).text.strip()
        data['cover'] = soup.select_one(self.details_cover_selector).get('src')

        if self.details_status_selector:
            if element := soup.select_one(self.details_status_selector):
                status = element.text.strip().lower()
                if status in ('completed', '完結済み'):
                    data['status'] = 'complete'
                elif status in ('ongoing', '連載中'):
                    data['status'] = 'ongoing'

        if self.details_authors_selector:
            if elements := soup.select(self.details_authors_selector):
                for element in elements:
                    # elements can be <a> or <p>
                    if element.name == 'p':
                        for author in element.text.strip().split(' - '):
                            data['authors'].append(author.strip())
                    else:
                        data['authors'].append(element.text.strip())

        if a_elements := soup.select(self.details_genres_selector):
            for a_element in a_elements:
                data['genres'].append(a_element.text.strip())

        if element := soup.select_one(self.details_synopsis_selector):
            data['synopsis'] = get_soup_element_inner_text(element, recursive=False)

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(data['slug'], soup=soup)

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for index, element in enumerate(soup.select('.page-chapter')):
            if not element.get('id') or not element.get('id').startswith('page'):
                continue

            image = None
            for attr in self.image_src_attrs:
                if image := element.img.get(attr):
                    break
            if image and image.startswith('//'):
                image = f'https:{image}'  # noqa: E231

            data['pages'].append(dict(
                slug=None,
                image=image,
                index=index + 1,
            ))

        return data

    def get_manga_chapters_data(self, slug, soup=None):
        """
        Returns manga chapters
        """
        def get_page(page, slug, soup=None):
            if soup is None and page is not None:
                r = self.session_get(
                    self.manga_url.format(slug),
                    params={
                        'page': page,
                    }
                )
                if r.status_code != 200:
                    return None

                mime_type = get_buffer_mime_type(r.content)
                if mime_type != 'text/html':
                    return None

                rtime = get_response_elapsed(r)
                soup = BeautifulSoup(r.text, 'lxml')
            else:
                rtime = 0

            items = []
            for li_element in soup.select(self.chapters_selector):
                if 'heading' in li_element.get('class'):
                    continue

                a_element = li_element.select_one('div a')
                url = a_element.get('href')
                slug = url.split('/')[-1]
                num = slug.split('-')[-1] if slug.startswith(f'{self.chapters_name}-') else None
                date = li_element.select_one('div:last-child').text.strip()

                items.append(dict(
                    slug=url.split('/', len(url.split('/')) - self.slug_segments)[-1],
                    title=a_element.text.strip(),
                    num=num if is_number(num) else None,
                    date=convert_date_string(date, self.date_format),
                ))

            if next_element := soup.select_one('.pagination a[rel="next"]'):
                next_url = next_element.get('href')
                next_page = int(parse_qs(urlparse(next_url).query)['page'][0])
                more = next_page > page
            else:
                more = False

            return items, more, rtime

        chapters = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page, slug, soup=soup)
            chapters += items
            soup = None

            delay = min(rtime * 4, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return list(reversed(chapters))

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200 and not self.ignore_images_status_code:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=f'{page["index"]:04d}.{mime_type.split("/")[-1]}' if page.get('index') else page['image'].split('/')[-1],  # noqa E231
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, orderby=None):
        if term:
            params = {
                'keyword': term,
            }
            url = self.search_url
            results_selector = self.results_selector
            last_chapter_selector = self.results_last_chapter_link_selector
        else:
            params = {}
            if orderby == 'populars':
                url = self.most_populars_url
                results_selector = self.results_selector
                last_chapter_selector = self.results_last_chapter_link_selector
            elif orderby == 'latest':
                url = self.latest_updates_url
                results_selector = self.results_latest_selector
                last_chapter_selector = self.results_last_chapter_lastest_updates_link_selector

        r = self.session.get(
            url,
            params=params,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select(results_selector):
            a_element = element.select_one(self.results_link_selector)
            img_element = element.select_one(self.results_cover_img_selector)
            if last_chapter_selector:
                last_chapter_element = element.select_one(last_chapter_selector)
            else:
                last_chapter_element = None

            name = a_element.get('title')
            if not name:
                name = a_element.text
            url = a_element.get('href')
            if last_chapter_element:
                last_chapter = last_chapter_element.text.replace(self.chapters_name.capitalize(), '').strip()
            else:
                last_chapter = None

            cover = None
            for attr in self.image_src_attrs:
                if cover := img_element.get(attr):
                    break

            results.append(dict(
                name=name.strip(),
                slug=url.split('/', len(url.split('/')) - self.slug_segments)[-1],
                cover=cover,
                last_chapter=last_chapter,
            ))

        return results

    @CompleteChallenge()
    def get_latest_updates(self):
        return self.get_manga_list(orderby='latest')

    @CompleteChallenge()
    def get_most_populars(self):
        return self.get_manga_list(orderby='populars')

    @CompleteChallenge()
    def search(self, term):
        return self.get_manga_list(term=term)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import time

import requests

from komikku.consts import USER_AGENT
from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed

SEARCH_RESULTS_PAGES = 10


class Perfscan(Server):
    id = 'perfscan'
    name = 'Perf Scan'
    lang = 'fr'

    base_url = 'https://perf-scan.xyz'
    logo_url = base_url + '/logo-static/Logo_Perf_NoText.png'
    api_url = 'https://api.perf-scan.xyz'
    manga_url = base_url + '/fr/series/{0}'
    api_manga_url = api_url + '/series/{0}'
    chapter_url = base_url + '/fr/series/{0}/chapter/{1}'
    api_chapter_url = api_url + '/series/{0}/chapter/{1}'
    media_url = api_url + '/cdn'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data using API requests

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Accept': 'application/json, text/plain, */*',
                'Origin': self.base_url,
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=resp_data.get('description'),
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = resp_data['title']
        data['cover'] = f'{self.media_url}/{resp_data["cover"]}'

        if resp_data.get('author'):
            data['authors'].append(resp_data['author'])
        if resp_data.get('artist'):
            data['authors'].append(resp_data['artist'])

        for genre in resp_data['SeriesGenre']:
            data['genres'].append(genre['Genre']['name'])
        if resp_data.get('Badge'):
            data['genres'].append(resp_data['Badge']['name'])

        if status := resp_data['Status']:
            if status['name'] == 'En cours':
                data['status'] = 'ongoing'
            elif status['name'] == 'Terminer':
                data['status'] = 'complete'
            elif status['name'] == 'Annulé':
                data['status'] = 'suspended'
            elif status['name'] == 'En Pause':
                data['status'] = 'hiatus'

        # Chapters
        for chapter in resp_data.get('Chapter'):
            title = chapter['title'].strip()
            if title in (None, '', '-'):
                title = f'Chapitre {chapter["index"]}'

            data['chapters'].append({
                'slug': chapter['id'],
                'title': title,
                'num': chapter.get('index'),
                'date': convert_date_string(chapter['createdAt'].split('T')[0], format='%Y-%m-%d'),
            })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are available in a <script> element
        """
        r = self.session_get(
            self.api_chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Accept': 'application/json, text/plain, */*',
                'Origin': self.base_url,
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        data = dict(
            pages=[],
        )
        for page in resp_data['content']:
            data['pages'].append({
                'slug': page['value'],
                'url': None,
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            f'{self.media_url}/{page["slug"]}',
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['slug'],
        )

    def get_manga_list(self, term=None, orderby=None):
        def get_page(page):
            params = dict(
                take=24,
                page=page,
                type='COMIC',
            )
            if term:
                params['title'] = term
            else:
                params.update(dict(
                    ranking='POPULAR',
                    rankingType='DAILY',
                ))

            r = self.session_get(
                self.api_url + '/series',
                params=params,
                headers={
                    'Accept': 'application/json, text/plain, */*',
                    'Origin': self.base_url,
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return None

            more = r.json()['data'] and page < SEARCH_RESULTS_PAGES

            return r.json()['data'], more, get_response_elapsed(r)

        results = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            for item in items:
                cover = item['thumbnail']
                if not cover.startswith('http'):
                    if self.media_url:
                        cover = f'{self.media_url}/{cover}'
                    else:
                        cover = f'{self.base_url}/{cover}'

                last_chapter = None
                if item.get('Chapter'):
                    last = item['Chapter'][-1]
                    last_chapter = f'Chapitre {last["index"]}'
                    if last.get('title') not in (None, '', '-'):
                        last_chapter = f'{last_chapter} : {last["title"]}'  # noqa: E203

                results.append(dict(
                    slug=item['id'],
                    name=item['title'],
                    cover=cover,
                    last_chapter=last_chapter,
                ))

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns most popular mangas
        """
        return self.get_manga_list(orderby='popular')

    def search(self, term):
        return self.get_manga_list(term=term)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging

from bs4 import BeautifulSoup

from komikku.servers.multi.heancms import HeanCMS
from komikku.servers.utils import parse_nextjs_hydration
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger(__name__)


class Aurorascans(HeanCMS):
    id = 'aurorascans'
    name = 'Aurora Scans'
    lang = 'en'
    status = 'disabled'  # Merged with Night Scans => Qi Scans

    base_url = 'https://aurorascans.com'
    logo_url = 'https://storage.aurorascans.com/public/upload/2025/01/26/JUST-TO-TRY-f17fe0b13caeba2c-16b03dbd7548f1a3.webp'
    api_url = 'https://api.aurorascans.com/api'

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are available in a <script> element
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if info := parse_nextjs_hydration(soup, 'images'):
            images = info[3]['children'][1][3]['children'][3]['children'][1][3]['children'][3]['children'][3]['chapter']['images']

            data = dict(
                pages=[],
            )
            for image in images:
                data['pages'].append(dict(
                    slug=None,
                    image=image['url'],
                ))

            return data

        return None



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.webview import CompleteChallenge


class Teamx(Server):
    id = 'teamx'
    name = 'Team-X'
    lang = 'ar'

    has_cf = True

    base_url = 'https://olympustaff.com'
    logo_url = base_url + '/assets/images/favicon.png'
    search_url = base_url + '/ajax/search'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/series/{0}/{1}'

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('.author-info-title h1').text.strip()
        data['cover'] = soup.select_one('.whitebox > .text-right > img').get('src')

        # Details
        data['genres'] = [a_element.text.strip() for a_element in soup.select('.review-author-info > a')]

        for element in soup.select('.whitebox > .text-right .full-list-info'):
            label = element.small.text.strip()
            a_elements = element.select('small:nth-child(2) a')

            if label.startswith('الحالة'):
                value = a_elements[0].text.strip()
                if value in ('قادم قريبًا', 'مستمرة'):
                    data['status'] = 'ongoing'
                elif value == 'مكتمل':
                    data['status'] = 'complete'
                elif value == 'متوقف':
                    data['status'] = 'suspended'

            elif label.startswith('الرسام'):
                for a_element in a_elements:
                    value = a_element.text.strip()
                    if value not in data['authors']:
                        data['authors'].append(value)

            elif label.startswith('النوع'):
                for a_element in a_elements:
                    value = a_element.text.strip()
                    data['genres'].append(value)

        # Synopsis
        data['synopsis'] = soup.select_one('.review-content > p').text.strip()

        # Chapters
        first_chapter_url = soup.select_one('.lastend .inepcx:nth-child(1) a').get('href')
        r = self.session_get(first_chapter_url)
        if r.status_code != 200:
            return data

        soup = BeautifulSoup(r.text, 'lxml')

        for option_element in reversed(soup.select('#select_chapter option')):
            chapter_url = option_element.get('value')
            if not chapter_url:
                continue

            slug = chapter_url.split('/')[-1]

            data['chapters'].append(dict(
                slug=slug,
                title=' '.join(option_element.text.strip().split()),
                num=slug if is_number(slug) else None,
                date=None,
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for img_element in soup.select('.image_list canvas[data-src], .image_list img[src]'):
            data['pages'].append(dict(
                image=img_element.get('src') or img_element.get('data-src'),
                slug=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.imgu a'):
            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=a_element.img.get('alt'),
                cover=a_element.img.get('src'),
            ))

        return results

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns most viewed mangas
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.swiper-slide .entry-image a'):
            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=a_element.img.get('alt'),
                cover=a_element.img.get('src'),
            ))

        return results

    @CompleteChallenge()
    def search(self, term):
        r = self.session_get(self.search_url, params=dict(keyword=term))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('a'):
            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=a_element.select_one('h4').text.strip(),
                cover=a_element.select_one('img').get('src'),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import logging

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

logger = logging.getLogger('komikku.support')


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/support.ui')
class SupportPage(Adw.NavigationPage):
    __gtype_name__ = 'SupportPage'

    title_box = Gtk.Template.Child('title_box')
    liberapay_button = Gtk.Template.Child('liberapay_button')
    paypal_button = Gtk.Template.Child('paypal_button')
    kofi_button = Gtk.Template.Child('kofi_button')
    bitcoin_label = Gtk.Template.Child('bitcoin_label')
    bitcoin_button = Gtk.Template.Child('bitcoin_button')

    payment_methods = {
        'liberapay': 'https://liberapay.com/valos/donate',
        'paypal': 'https://www.paypal.com/donate?business=GSRGEQ78V97PU&no_recurring=0&item_name=You+can+help+me+to+keep+developing+apps+through+donations.&currency_code=EUR',
        'ko-fi': 'https://ko-fi.com/X8X06EM3L',
        'bitcoin': 'bc1qzq2ve4c44hu7kzgjth3gmq0tnyl33eltlvudcw',
    }

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window

        self.window.breakpoint.add_setter(self.title_box, 'orientation', Gtk.Orientation.VERTICAL)

        self.liberapay_button.connect('clicked', self.on_button_clicked, 'liberapay')
        self.paypal_button.connect('clicked', self.on_button_clicked, 'paypal')
        self.kofi_button.connect('clicked', self.on_button_clicked, 'ko-fi')
        self.bitcoin_label.set_text(self.payment_methods['bitcoin'])
        self.bitcoin_button.connect('clicked', self.on_button_clicked, 'bitcoin')

    def on_button_clicked(self, button, method):
        if method == 'bitcoin':
            clipboard = Gdk.Display.get_clipboard(Gdk.Display.get_default())
            content = Gdk.ContentProvider.new_for_bytes('text/plain', GLib.Bytes.new(bytes(self.payment_methods[method], 'utf-8')))
            clipboard.set_content(content)
        else:
            if uri := self.payment_methods.get(method):
                Gtk.UriLauncher.new(uri=uri).launch()

    def show(self):
        self.window.navigationview.push(self)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Lelscan(Server):
    id = 'lelscan'
    name = 'Lelscan'
    lang = 'fr'
    true_search = False

    base_url = 'https://lelscans.net'
    logo_url = base_url + '/images/icones/favicon.ico'
    most_populars_url = base_url + '/lecture-en-ligne-one-piece'
    chapter_url = base_url + '/scan-{0}/{1}'
    page_url = base_url + '/scan-{0}/{1}/{2}'
    cover_url = base_url + '/mangas/{0}/thumb_cover.jpg'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        r = self.session_get(initial_data['url'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status='ongoing',
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.select_one('.outil_lecture h2').text.split('-')[0].strip()
        data['cover'] = self.cover_url.format(data['slug'])

        # Chapters
        for option_element in reversed(soup.select('#header-image select:first-child option')):
            url = option_element.get('value')
            slug = url.split('/')[-1]

            data['chapters'].append(dict(
                slug=slug,
                title='Chapitre {0}'.format(slug),
                num=slug if is_number(slug) else None,
                date=None,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for a_element in soup.select('#navigation a'):
            text = a_element.text.strip()
            if not text.isdigit():
                continue

            slug = a_element.get('href').split('/')[-1]
            data['pages'].append(dict(
                slug=slug,
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page content
        """
        r = self.session_get(self.page_url.format(manga_slug, chapter_slug, page['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        path = soup.select_one('#image img').get('src')

        r = self.session_get(self.base_url + path)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=path.split('/')[-1].split('?')[0],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return url

    def get_latest_updates(self):
        """
        Returns hot manga
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = []
        for a_element in soup.select('#main_hot_ul a.hot_manga_img'):
            url = a_element.get('href')

            slug = url.split('/')[-1]
            for chunk in ('lecture-ligne-', 'lecture-en-ligne-', '.php'):
                slug = slug.replace(chunk, '')

            data.append(dict(
                slug=slug,
                name=a_element.get('title').replace('Scan', '').strip(),
                cover=self.cover_url.format(slug),
                url=url,
            ))

        return data

    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = []
        for option_element in soup.select('#header-image select:last-child option'):
            url = option_element.get('value')

            slug = url.split('/')[-1]
            for chunk in ('lecture-ligne-', 'lecture-en-ligne-', '.php'):
                slug = slug.replace(chunk, '')

            data.append(dict(
                slug=slug,
                name=option_element.text.strip(),
                cover=self.cover_url.format(slug),
                url=url,
            ))

        return data

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
import datetime
from gettext import gettext as _
from gettext import ngettext
import threading
import time

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.models import Chapter
from komikku.models import create_db_connection
from komikku.models import Download
from komikku.models import insert_rows
from komikku.models import Settings
from komikku.utils import if_network_available


class Downloader(GObject.GObject):
    """
    Chapters downloader
    """
    __gsignals__ = {
        'download-changed': (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, )),
        'ended': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'started': (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    current_download = None
    running = False
    stop_flag = False

    def __init__(self, window):
        GObject.GObject.__init__(self)

        self.window = window

    def add(self, chapters, emit_signal=False):
        """Add chapters to download"""
        chapters_ids = []
        rows_data = []
        db_conn = create_db_connection()

        for chapter in chapters:
            if isinstance(chapter, Chapter):
                if chapter.downloaded:
                    continue
                chapter_id = chapter.id
            else:
                chapter_id = chapter

            if Download.get_by_chapter_id(chapter_id, db_conn=db_conn) is not None:
                # Chapter download is already scheduled
                continue

            rows_data.append(dict(
                chapter_id=chapter_id,
                status='pending',
                percent=0,
                date=datetime.datetime.now(datetime.UTC),
            ))
            chapters_ids.append(chapter_id)

        if not chapters_ids:
            return

        with db_conn:
            insert_rows(db_conn, 'downloads', rows_data)

        if emit_signal:
            for chapter_id in chapters_ids:
                download = Download.get_by_chapter_id(chapter_id, db_conn=db_conn)
                if download:
                    self.emit('download-changed', download, None)

    def cancel(self, chapters):
        """Cancel downloads"""
        db_conn = create_db_connection()

        for chapter in chapters:
            download = Download.get_by_chapter_id(chapter.id, db_conn=db_conn)
            if not download:
                continue

            if self.current_download and self.current_download.id == download.id:
                self.current_download.status = 'canceled'

            download.delete(db_conn=db_conn)

            self.emit('download-changed', None, chapter)

    @if_network_available
    def start(self):
        def run(exclude_errors=False):
            db_conn = create_db_connection()

            # Get pending downloads sorted by server
            sql = """
                SELECT d.id, m.id AS manga_id, m.server_id FROM downloads d
                JOIN chapters c ON d.chapter_id = c.id
                JOIN mangas m ON c.manga_id = m.id
            """
            if exclude_errors:
                sql += ' WHERE d.status != "error"'

            sql += ' ORDER BY m.server_id ASC, m.id ASC, d.id ASC'
            rows = db_conn.execute(sql).fetchall()

            if not rows or self.stop_flag:
                self.running = False
                GLib.idle_add(self.emit, 'ended')
                return

            # Build dict with list of downloads per server
            server_downloads = {}
            for row in rows:
                if row['server_id'] not in server_downloads:
                    server_downloads[row['server_id']] = {}
                if row['manga_id'] not in server_downloads[row['server_id']]:
                    server_downloads[row['server_id']][row['manga_id']] = []
                server_downloads[row['server_id']][row['manga_id']].append(row['id'])

            with ThreadPoolExecutor(max_workers=len(server_downloads)) as executor:
                tasks = {}
                for _server_id, manga_downloads in server_downloads.items():
                    for _manga_id, downloads in manga_downloads.items():
                        future = executor.submit(process_manga_downloads, downloads)
                        tasks[future] = None

                for _future in as_completed(tasks):
                    if self.stop_flag:
                        executor.shutdown(False, cancel_futures=True)
                        break

            # Continue, new downloads may have been added in the meantime
            run(exclude_errors=True)

        def process_manga_downloads(downloads):
            db_conn = create_db_connection()
            manga = None
            cancelations = 0
            errors = 0
            failures = 0
            successes = 0

            for download_id in downloads:
                self.current_download = None
                if self.stop_flag:
                    break

                download = Download.get(download_id, db_conn=db_conn)
                if download is None:
                    # Download has been canceled in the meantime
                    cancelations += 1
                    continue

                chapter = download.chapter
                if manga is None:
                    manga = chapter.manga

                download.update(dict(status='downloading'), db_conn=db_conn)
                GLib.idle_add(self.emit, 'download-changed', download, None)

                try:
                    if chapter.update_full(db_conn=db_conn) and len(chapter.pages) > 0:
                        self.current_download = download
                        page_errors = 0
                        page_successes = 0

                        for index, _page in enumerate(chapter.pages):
                            if self.stop_flag:
                                # Downloader has been stopped
                                break
                            if download.status == 'canceled':
                                # Download has been canceled
                                break

                            if chapter.get_page_path(index) is None:
                                # Depending on amount of bandwidth the server has, we must be mindful not to overload it
                                # with our requests.
                                #
                                # Furthermore, multiple and fast-paced requests from same IP address can alert the system
                                # administrator that potentially unwanted actions are taking place.
                                # This may result in an IP ban.
                                #
                                # Easiest way to avoid server  overloading is to set a time-out between requests
                                # equal to 2x the time it took to load the page (responsive delay).
                                path, rtime = chapter.get_page(index, db_conn=db_conn)

                                if path is not None:
                                    page_successes += 1
                                    download.update(dict(percent=(index + 1) * 100 / len(chapter.pages)), db_conn=db_conn)
                                else:
                                    page_errors += 1
                                    download.update(dict(errors=page_errors), db_conn=db_conn)

                                GLib.idle_add(self.emit, 'download-changed', download, None)

                                if index < len(chapter.pages) - 1 and not self.stop_flag and rtime:
                                    time.sleep(min(2 * rtime, DOWNLOAD_MAX_DELAY))
                            else:
                                page_successes += 1

                        if page_successes == len(chapter.pages):
                            # Download ended
                            # All pages were successfully downloaded
                            successes += 1
                            chapter.update(dict(downloaded=1), db_conn=db_conn)
                            download.delete(db_conn=db_conn)
                            GLib.idle_add(self.emit, 'download-changed', None, chapter)

                        elif page_successes + page_errors == len(chapter.pages):
                            # Download has ended
                            # But at least one page failed to be downloaded
                            errors += 1
                            download.update(dict(status='error'), db_conn=db_conn)
                            GLib.idle_add(self.emit, 'download-changed', download, None)

                        else:
                            # Download has been interruped
                            if self.stop_flag:
                                # Downloader has been stopped
                                # Restore pending status of current unfinished download
                                download.update(dict(status='pending'), db_conn=db_conn)

                            elif download.status == 'canceled':
                                # Download has been canceled
                                cancelations += 1
                                GLib.idle_add(self.emit, 'download-changed', None, chapter)
                    else:
                        # Possible causes:
                        # - Empty chapter
                        # - Outdated chapter info
                        # - Server has undergone changes (API, HTML) and plugin code is outdated
                        failures += 1
                        download.update(dict(status='error'), db_conn=db_conn)
                        GLib.idle_add(self.emit, 'download-changed', download, None)
                except Exception:
                    # Possible causes:
                    # - No Internet connection
                    # - Connection timeout, read timeout
                    # - Server down
                    # - Bad/corrupt local archive
                    failures += 1
                    download.update(dict(status='error'), db_conn=db_conn)
                    GLib.idle_add(self.emit, 'download-changed', download, None)

            self.current_download = None
            if not self.stop_flag:
                GLib.idle_add(
                    notify_manga_chapters_download_ended,
                    manga, len(downloads), successes, errors, cancelations, failures
                )

        def notify_manga_chapters_download_ended(manga, total, successes, errors, cancelations, failures):
            if notification is None:
                return

            if cancelations + errors + failures == 0:
                notification.set_category('transfer.success')
            elif errors + failures != 0:
                notification.set_category('transfer.error')

            notification.set_title(_('[{0}] Chapters download ended').format(manga.name))

            body = []
            if successes:
                body.append('🟢 ' + '{0}/{1} completed'.format(successes, total))
            if cancelations:
                body.append('🟡 ' + '{0}/{1} canceled'.format(cancelations, total))
            if errors:
                body.append('🟠 ' + '{0}/{1} with errors'.format(errors, total))
            if failures:
                body.append('🔴 ' + '{0}/{1} failed'.format(failures, total))

            notification.set_body(' | '.join(body))
            self.window.application.send_notification(f'downloader.{manga.id}', notification)

        if self.running:
            return

        Settings.get_default().downloader_state = True
        self.running = True
        self.stop_flag = False

        if Settings.get_default().desktop_notifications:
            notification = Gio.Notification()
        else:
            notification = None

        GLib.idle_add(self.emit, 'started')

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def stop(self, save_state=False):
        if not self.running:
            return

        self.stop_flag = True
        if save_state:
            Settings.get_default().downloader_state = False


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/download_manager.ui')
class DownloadManagerPage(Adw.NavigationPage):
    __gtype_name__ = 'DownloadManagerPage'
    __gsignals_handlers_ids__ = None

    selection_mode = False
    selection_mode_range = False
    selection_mode_last_row_index = None

    left_button = Gtk.Template.Child('left_button')
    title = Gtk.Template.Child('title')
    start_stop_button = Gtk.Template.Child('start_stop_button')
    menu_button = Gtk.Template.Child('menu_button')

    stack = Gtk.Template.Child('stack')
    listbox = Gtk.Template.Child('listbox')
    selection_mode_actionbar = Gtk.Template.Child('selection_mode_actionbar')

    def __init__(self, window):
        Adw.NavigationPage.__init__(self)

        self.window = window
        self.downloader = self.window.downloader

        self.builder = window.builder
        self.builder.add_from_resource('/info/febvre/Komikku/ui/menu/download_manager.xml')

        # Header bar
        self.left_button.connect('clicked', self.leave_selection_mode)
        self.start_stop_button.connect('clicked', self.on_start_stop_button_clicked)
        self.menu_button.set_menu_model(self.builder.get_object('menu-download-manager'))
        # Focus is lost after showing popover submenu (bug?)
        self.menu_button.get_popover().connect('closed', lambda _popover: self.menu_button.grab_focus())

        self.listbox.connect('row-activated', self.on_download_row_activated)
        self.listbox.connect('selected-rows-changed', self.on_selection_changed)
        self.window.controller_key.connect('key-pressed', self.on_key_pressed)

        # Gestures for multi-selection mode
        self.gesture_click = Gtk.GestureClick.new()
        self.gesture_click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_click.set_button(3)
        self.gesture_click.connect('pressed', self.on_download_row_right_click)
        self.listbox.add_controller(self.gesture_click)

        self.gesture_long_press = Gtk.GestureLongPress.new()
        self.gesture_long_press.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture_long_press.set_touch_only(False)
        self.gesture_long_press.connect('pressed', self.on_gesture_long_press_activated)
        self.listbox.add_controller(self.gesture_long_press)

        self.__gsignals_handlers_ids__ = [
            self.downloader.connect('download-changed', self.update_row),
            self.downloader.connect('ended', self.update_headerbar),
            self.downloader.connect('started', self.update_headerbar),
        ]

        self.window.navigationview.add(self)

    def add_actions(self):
        # Delete All action
        delete_all_action = Gio.SimpleAction.new('download-manager.delete-all', None)
        delete_all_action.connect('activate', self.on_menu_delete_all_clicked)
        self.window.application.add_action(delete_all_action)

        # Delete Selected action
        delete_selected_action = Gio.SimpleAction.new('download-manager.delete-selected', None)
        delete_selected_action.connect('activate', self.on_menu_delete_selected_clicked)
        self.window.application.add_action(delete_selected_action)

    def enter_selection_mode(self):
        self.props.can_pop = False
        self.left_button.set_label(_('Cancel'))
        self.left_button.set_tooltip_text(_('Cancel'))
        self.left_button.set_visible(True)
        self.start_stop_button.set_visible(False)
        self.menu_button.set_visible(False)

        self.selection_mode = True

        self.listbox.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
        self.selection_mode_actionbar.set_revealed(True)

    def leave_selection_mode(self, *args):
        self.props.can_pop = True
        self.left_button.set_visible(False)
        self.start_stop_button.set_visible(True)
        self.menu_button.set_visible(True)

        self.selection_mode = False

        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        for row in self.listbox:
            row._selected = False
        self.selection_mode_actionbar.set_revealed(False)

    def on_download_row_activated(self, _listbox, row):
        row.grab_focus()

        if not self.selection_mode:
            return

        if self.selection_mode_range and self.selection_mode_last_row_index is not None:
            # Range selection mode: select all rows between last selected row and clicked row
            walk_index = self.selection_mode_last_row_index
            last_index = row.get_index()

            while walk_index != last_index:
                walk_row = self.listbox.get_row_at_index(walk_index)
                if walk_row and not walk_row._selected:
                    self.listbox.select_row(walk_row)
                    walk_row._selected = True

                if walk_index < last_index:
                    walk_index += 1
                else:
                    walk_index -= 1

        self.selection_mode_range = False

        if row._selected:
            self.listbox.unselect_row(row)
            self.selection_mode_last_row_index = None
            row._selected = False
        else:
            self.listbox.select_row(row)
            self.selection_mode_last_row_index = row.get_index()
            row._selected = True

        if len(self.listbox.get_selected_rows()) == 0:
            self.leave_selection_mode()

    def on_download_row_right_click(self, _gesture, _n_press, _x, y):
        """Allow to enter in selection mode with a right click on a row"""
        if self.selection_mode:
            return Gdk.EVENT_PROPAGATE

        row = self.listbox.get_row_at_y(y)
        if not self.selection_mode and row is not None:
            self.enter_selection_mode()
            self.on_download_row_activated(None, row)
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def on_gesture_long_press_activated(self, _gesture, _x, y):
        """Allow to enter in selection mode with a long press on a row"""
        if not self.selection_mode:
            self.enter_selection_mode()
        else:
            # Enter in 'Range' selection mode
            # Long press on a download row then long press on another to select everything in between
            self.selection_mode_range = True

        selected_row = self.listbox.get_row_at_y(y)
        self.on_download_row_activated(None, selected_row)

    def on_key_pressed(self, _controller, keyval, _keycode, state):
        if self.window.page != self.props.tag:
            return Gdk.EVENT_PROPAGATE

        modifiers = state & Gtk.accelerator_get_default_mod_mask()

        if self.selection_mode:
            if keyval == Gdk.KEY_Escape or (modifiers == Gdk.ModifierType.ALT_MASK and keyval in (Gdk.KEY_Left, Gdk.KEY_KP_Left)):
                self.leave_selection_mode()
                # Stop event to prevent back navigation
                return Gdk.EVENT_STOP
        else:
            # Allow to enter in selection mode with <SHIFT>+Arrow key
            if modifiers != Gdk.ModifierType.SHIFT_MASK or keyval not in (Gdk.KEY_Up, Gdk.KEY_KP_Up, Gdk.KEY_Down, Gdk.KEY_KP_Down):
                return Gdk.EVENT_PROPAGATE

            if row := self.listbox.get_focus_child():
                self.enter_selection_mode()
                self.on_download_row_activated(None, row)

        return Gdk.EVENT_PROPAGATE

    def on_menu_delete_all_clicked(self, _action, _param):
        chapters = []

        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            chapters.append(row.download.chapter)
            self.listbox.remove(row)
            row = next_row

        self.downloader.cancel(chapters)

        self.leave_selection_mode()
        self.update_headerbar()
        GLib.idle_add(self.stack.set_visible_child_name, 'empty')

    def on_menu_delete_selected_clicked(self, _action, _param):
        chapters = []

        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            if row._selected:
                chapters.append(row.download.chapter)
                self.listbox.remove(row)
            row = next_row

        self.downloader.cancel(chapters)

        self.leave_selection_mode()
        self.update_headerbar()
        if self.listbox.get_first_child() is None:
            # No more downloads
            GLib.idle_add(self.stack.set_visible_child_name, 'empty')

    def on_selection_changed(self, _flowbox):
        number = len(self.listbox.get_selected_rows())
        if number:
            self.title.set_subtitle(ngettext('{0} selected', '{0} selected', number).format(number))
        else:
            self.title.set_subtitle('')

    @if_network_available
    def on_start_stop_button_clicked(self, _button):
        self.start_stop_button.set_sensitive(False)

        if self.downloader.running:
            self.downloader.stop(save_state=True)
        else:
            self.downloader.start()

    def populate(self):
        # Clear
        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()
            self.listbox.remove(row)
            row = next_row

        db_conn = create_db_connection()
        records = db_conn.execute('SELECT * FROM downloads ORDER BY date ASC').fetchall()

        if records:
            for record in records:
                download = Download.get(record['id'], db_conn=db_conn)

                row = DownloadRow(download)
                self.listbox.append(row)

            self.stack.set_visible_child_name('list')
        else:
            # No downloads
            self.stack.set_visible_child_name('empty')

    def select_all(self):
        if not self.selection_mode:
            self.enter_selection_mode()

        for row in self.listbox:
            if row._selected:
                continue
            self.listbox.select_row(row)
            row._selected = True

    def show(self):
        self.populate()

        self.update_headerbar(forced=True)
        self.window.navigationview.push(self)

    def update_headerbar(self, *args, forced=False):
        if self.window.page != self.props.tag and not forced:
            return

        if self.listbox.get_first_child() is not None:
            if self.downloader.running:
                self.start_stop_button.get_first_child().set_from_icon_name('media-playback-stop-symbolic')
                self.menu_button.set_visible(False)
            else:
                self.start_stop_button.get_first_child().set_from_icon_name('media-playback-start-symbolic')
                self.menu_button.set_visible(True)

            self.start_stop_button.set_sensitive(True)
            self.start_stop_button.set_visible(True)
        else:
            # No downloads
            self.start_stop_button.set_visible(False)
            self.menu_button.set_visible(False)

    def update_row(self, _downloader, download, chapter):
        chapter_id = chapter.id if chapter is not None else download.chapter.id

        for row in self.listbox:
            if row.download.chapter.id == chapter_id:
                row.download = download
                if row.download:
                    row.update()
                else:
                    self.listbox.remove(row)
                break

        if self.listbox.get_first_child() is None:
            # No more downloads
            self.stack.set_visible_child_name('empty')


class DownloadRow(Gtk.ListBoxRow):
    _selected = False

    def __init__(self, download):
        Gtk.ListBoxRow.__init__(self)

        self.add_css_class('download-manager-download-listboxrow')

        self.download = download

        if self.download.percent and download.chapter.pages:
            nb_pages = len(download.chapter.pages)
            counter = int((nb_pages / 100) * self.download.percent)
            fraction = self.download.percent / 100
        else:
            counter = None
            fraction = None

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        # Manga
        label = Gtk.Label(xalign=0)
        label.add_css_class('body')
        label.set_valign(Gtk.Align.CENTER)
        label.set_wrap(True)
        label.set_text(download.chapter.manga.name)
        hbox.append(label)

        # Progress label
        self.progress_label = Gtk.Label(xalign=0)
        self.progress_label.add_css_class('caption')
        self.progress_label.set_valign(Gtk.Align.CENTER)
        self.progress_label.set_wrap(True)
        text = _(Download.STATUSES[self.download.status]).upper() if self.download.status == 'error' else ''
        if counter:
            text = f'{text} {counter}/{nb_pages}'
        if text:
            self.progress_label.set_text(text)
        hbox.append(self.progress_label)

        vbox.append(hbox)

        # Chapter
        label = Gtk.Label(xalign=0)
        label.add_css_class('caption')
        label.set_valign(Gtk.Align.CENTER)
        label.set_wrap(True)
        label.set_text(download.chapter.title)
        vbox.append(label)

        # Progress bar
        self.progressbar = Gtk.ProgressBar()
        self.progressbar.set_show_text(False)
        if fraction:
            self.progressbar.set_fraction(fraction)
        vbox.append(self.progressbar)

        self.set_child(vbox)

    def update(self):
        """
        Updates chapter download progress
        """
        if not self.download.chapter.pages:
            return

        nb_pages = len(self.download.chapter.pages)
        counter = int((nb_pages / 100) * self.download.percent)
        fraction = self.download.percent / 100

        self.progressbar.set_fraction(fraction)
        text = _(Download.STATUSES[self.download.status]).upper() if self.download.status == 'error' else ''
        text = f'{text} {counter}/{nb_pages}'
        self.progress_label.set_text(text)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import json

from bs4 import BeautifulSoup

from komikku.servers.multi.madara import Madara
from komikku.webview import CompleteChallenge
from komikku.utils import get_buffer_mime_type


class Manhwahentai(Madara):
    id = 'manhwahentai'
    name = 'Manhwa Hentai'
    lang = 'en'
    is_nsfw_only = True
    status = 'disabled'

    date_format = '%d %B %Y'
    series_name = 'pornhwa'

    base_url = 'https://manhwahentai.to'
    chapters_url = base_url + '/pornhwa/{0}/ajax/chapters/'

    chapters_list_selector = '.manga-post-chapters'
    details_name_selector = '.post-title p.h1'
    details_authors_selector = '.post-tax-wp-manga-artist .post-tags .tag-name'
    details_genres_selector = '.post-tax-wp-manga-category .post-tags .tag-name'
    details_status_selector = None
    details_synopsis_selector = '.summary_content .post-meta:-soup-contains("Description") .post-tag .tag-name'
    results_selector = '.manga'
    result_name_slug_selector = '.post-title a'
    result_cover_selector = '.item-thumb img'

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        # Pages images are loaded via javascript
        for script_element in soup.find_all('script'):
            script = script_element.string
            if script is None or script_element.get('id') != 'chapter_preloaded_images':
                continue

            for line in script.split('\n'):
                if 'chapter_preloaded_images' not in line:
                    continue
                line = line.strip()
                json_data = line.split('=')[-1]
                if json_data[-2] == ',':
                    json_data = json_data[:-2] + ']'
                for image in json.loads(json_data):
                    data['pages'].append(dict(
                        slug=None,
                        image=image['src'],
                    ))
                break

        return data



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
import time
from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests
from urllib.parse import urlsplit

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.consts import USER_AGENT_MOBILE
from komikku.servers import Server
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number

CHAPTERS_PER_REQUEST = 30
LANGUAGES_CODES = dict(
    en='en',
    es='es',
    fr='fr',
    id='id',
    th='th',
    zh_Hant='zh-hant',  # diff
)

SERVER_NAME = 'WEBTOON'


class Webtoon(Server):
    id = 'webtoon'
    name = SERVER_NAME
    lang = 'en'
    status = 'disabled'  # commercial platforms are no supported

    base_url = 'https://www.webtoons.com'
    logo_url = 'https://webtoons-static.pstatic.net/image/favicon/favicon.ico?dt=2017082301'
    search_url = base_url + '/{0}/search/{1}'
    most_populars_url = base_url + '/{0}/ranking/{1}'
    manga_url = base_url + '{0}'
    api_chapters_url = 'https://m.webtoons.com/api/v1/{type}/{sid}/episodes'
    chapter_url = base_url + '{0}'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': 'all',
            'options': [
                {'key': 'all', 'name': _('All')},
                {'key': 'originals', 'name': _('Originals')},
                {'key': 'canvas', 'name': _('Canvas')},
            ],
        },
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        return dict(url=url.replace(cls.base_url, ''), slug=url.split('=')[-1])

    def _get_manga_cover_url(self, url):
        # No cover in manga page, use RSS feed
        r = self.session_get(self.manga_url.format(url.replace('list?', 'rss?')), headers={'User-Agent': USER_AGENT})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/xml':
            return None

        soup = BeautifulSoup(r.text, features='xml')
        if element := soup.select_one('image url'):
            return element.text.strip()

        return None

    def _parse_results_page(self, response):
        soup = BeautifulSoup(response.text, 'lxml')

        results = []
        for a_element in soup.select('.webtoon_list > li > a'):
            # Small difference here compared to the majority of servers
            # slug can't be used to forge manga URL, we must store the full url (relative)
            results.append(dict(
                slug=a_element.get('href').split('=')[-1],
                url=a_element.get('href').replace(self.base_url, ''),
                name=a_element.select_one('.info_text .title').text.strip(),
                cover=a_element.select_one('img').get('src'),
            ))

        return results

    def _search_by_ranking(self, ranking):
        # Clear cookies
        # Seems to help to bypass some region-based restrictions?!?
        self.session.cookies.clear()

        r = self.session_get(
            self.most_populars_url.format(LANGUAGES_CODES[self.lang], ranking),
            headers={
                'User-Agent': USER_AGENT,
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        return self._parse_results_page(r)

    def _search_by_type(self, term, type):
        # Clear cookies
        # Seems to help to bypass some region-based restrictions?!?
        self.session.cookies.clear()

        r = self.session_get(
            self.search_url.format(LANGUAGES_CODES[self.lang], type),
            params=dict(
                keyword=term,
                searchMode='ALL',
            ),
            headers={'User-Agent': USER_AGENT}
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        return self._parse_results_page(r)

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's url (provided by search)
        """
        assert 'url' in initial_data, 'Manga url is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['url']), headers={'User-Agent': USER_AGENT})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        # Get true URL after redirects
        split_url = urlsplit(r.url)
        url = '{0}?{1}'.format(split_url.path, split_url.query)

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            url=url,
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = get_soup_element_inner_text(soup.find(class_='subj'))

        # Details
        info_element = soup.find('div', class_='info')
        for element in info_element.find_all(class_='genre'):
            data['genres'].append(get_soup_element_inner_text(element))

        if 'canvas' in data['url']:
            # Canvas/Challenge
            detail_element = soup.find('div', class_='detail')

            data['cover'] = soup.find('div', class_='detail_header').img.get('src')

            for element in info_element.find_all(class_='author'):
                data['authors'].append(get_soup_element_inner_text(element))
        else:
            # Original/Webtoon
            detail_element = soup.find('div', class_='detail_body')

            data['cover'] = self._get_manga_cover_url(data['url'])

            try:
                for element in soup.find('div', class_='_authorInnerContent').find_all('h3'):
                    data['authors'].append(element.text.strip())
            except Exception:
                for element in info_element.find_all(class_='author'):
                    data['authors'].append(get_soup_element_inner_text(element))

            status_class = ''.join(detail_element.find('p', class_='day_info').span.get('class'))
            if 'completed' in status_class:
                data['status'] = 'complete'
            else:
                data['status'] = 'ongoing'

        data['synopsis'] = detail_element.find('p', class_='summary').text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(data['url'])

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_url), headers={'User-Agent': USER_AGENT})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        imgs = soup.find('div', id='_imageList').find_all('img')

        data = dict(
            pages=[],
        )
        for index, img in enumerate(imgs):
            data['pages'].append(dict(
                index=index + 1,
                slug=None,  # slug can't be used to forge image URL
                image=img.get('data-url').strip(),
            ))

        return data

    def get_manga_chapters_data(self, url):
        """
        Returns manga chapters data using API of mobile site
        """
        chapters = []
        sid = parse_qs(urlparse(url).query)['title_no'][0]
        if '/canvas/' in url:
            type = 'canvas'
        else:
            type = 'webtoon'

        def get_page(cursor):
            params = dict(
                pageSize=CHAPTERS_PER_REQUEST,
            )
            if cursor:
                params['cursor'] = cursor

            r = self.session_get(
                self.api_chapters_url.format(type=type, sid=sid),
                params=params,
                headers={
                    'User-Agent': USER_AGENT_MOBILE,
                }
            )
            if r.status_code != 200:
                return None

            next_cursor = r.json()['result']['nextCursor']

            return r.json()['result']['episodeList'], next_cursor, get_response_elapsed(r)

        chapters = []
        cursor = None
        delay = None
        while cursor != 0:
            if delay:
                time.sleep(delay)

            episodes, cursor, rtime = get_page(cursor)
            for episode in episodes:
                title = episode['episodeTitle']
                if episode.get('hasBgm'):
                    title += ' ♫'
                num = episode['episodeNo']

                chapters.append(dict(
                    slug=str(num),  # slug nust be a string
                    title=title,
                    num=num if is_number(num) else None,
                    date=datetime.datetime.fromtimestamp(episode['exposureDateMillis'] / 1000).date(),
                    url=episode['viewerLink'],
                ))

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None

        return chapters

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'], headers={'Referer': self.base_url, 'User-Agent': USER_AGENT})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(url)

    def get_most_populars(self, type='all'):
        """
        Returns popular manga
        """
        results = None

        if type == 'all' or type == 'originals':
            if originals_results := self._search_by_ranking('popular'):
                results = originals_results

        if type == 'all' or type == 'canvas':
            if canvas_results := self._search_by_ranking('canvas'):
                if results is None:
                    results = canvas_results
                else:
                    results += canvas_results

        return results

    def is_long_strip(self, _manga_data):
        return True

    def search(self, term, type='all'):
        results = None

        if type == 'all' or type == 'originals':
            if originals_results := self._search_by_type(term, 'originals'):
                results = originals_results

        if type == 'all' or type == 'canvas':
            if canvas_results := self._search_by_type(term, 'canvas'):
                if results is None:
                    results = canvas_results
                else:
                    results += canvas_results

        return results


class Dongmanmanhua(Webtoon):
    id = 'dongmanmanhua:webtoon'
    name = 'Dongman Manhua'
    lang = 'zh_Hans'

    base_url = 'https://www.dongmanmanhua.cn'
    search_url = base_url + '/search'
    most_populars_url = base_url + '/top'
    manga_url = base_url + '{0}'
    chapters_url = 'https://m.dongmanmanhua.cn/{0}'
    chapter_url = base_url + '{0}'


class Webtoon_es(Webtoon):
    id = 'webtoon_es'
    name = SERVER_NAME
    lang = 'es'


class Webtoon_fr(Webtoon):
    id = 'webtoon_fr'
    name = SERVER_NAME
    lang = 'fr'


class Webtoon_id(Webtoon):
    id = 'webtoon_id'
    name = SERVER_NAME
    lang = 'id'


class Webtoon_th(Webtoon):
    id = 'webtoon_th'
    name = SERVER_NAME
    lang = 'th'


class Webtoon_zh_hant(Webtoon):
    id = 'webtoon_zh_hant'
    name = SERVER_NAME
    lang = 'zh_Hant'
